#ifndef BPPDEF_H
#define BPPDEF_H

#include <limits.h>
#include <stddef.h>

#ifdef __GNUC__
#ifndef GCC
#define GCC
#endif
#endif

#ifdef GCC
  #ifndef LLONG_MAX
	 #define LLONG_MAX __LONG_LONG_MAX__
  #endif
  #ifndef INT_MAX
	 #define INT_MAX __INT_MAX__
  #endif
  #ifndef SHRT_MAX
	 #define SHRT_MAX __SHRT_MAX__
  #endif
  #ifndef UINT_MAX
	 #define UINT_MAX __UINT_MAX__
  #endif
  #ifndef USHRT_MAX
	 #define USHRT_MAX __USHRT_MAX__
  #endif
  #ifndef SIZE_MAX
	 #if __WORDSIZE == 64
		#define SIZE_MAX UINT64_MAX
	#else
		#define SIZE_MAX UINT32_MAX
	#endif
  #endif
#endif
//#define EMULATE_I64
#ifndef WINCE
  #ifdef _WIN32_WCE
	 #define WINCE
  #endif
#endif

#ifndef LINUX
#if defined(__linux__) || defined(__linux)
#define LINUX
#endif
#endif

#ifndef UNIX
#if defined(LINUX) || defined(__unix__) || defined(__unix) || (defined(GCC) && !defined(WIN32))
#define UNIX
#endif
#endif

#ifdef _MSC_VER
  #if _MSC_VER <= 800 
	  typedef int bool;
	  #define true 1
	  #define false 0
	  #define __NOSTL__
	  #define NOCPPEXCEPTION 
	  #define asm __asm
  #endif
  #if _MSC_VER >= 1900 
	 #define HAVE_THREADLOCALE
  #endif
#endif
#ifndef NOASM
  #ifdef WINCE
	  #define NOASM
  #endif
  #ifdef GCC
	  #define NOASM
 //    #define __NOSTL__
  #endif
  #ifdef _WIN64
	 #define NOASM
	#ifndef WIN32
	  #define WIN32
	#endif
  #endif
#endif
#ifdef WINCE
// on Windows CE WIN32 must be defined too
#ifndef WIN32
  #define WIN32
#endif
	 typedef int ptrdiff_t;
	#define EDOM 33
	#define ERANGE 34
	#undef STRSAFE_NO_DEPRECATE
	#define DEPRECATE_SUPPORTED  
	// not all versions of WinCE SDK support C++ exceptions
  #define NOCPPEXCEPTION
  // CRT library for WinCE doesn't support conio.h functions
  #define NOCONIO
  // not all versions of WinCE SDK support STL
  #define __NOSTL__
#endif

#ifdef WIN32
	#ifndef _CONSOLE
		// console io is useable only in console applications
		#ifndef NOCONIO
			#define NOCONIO
		#endif
	#endif
  #include <windows.h>
  #ifdef BUILD_BPPLIB
	 #define BPPLIB_DLLEXP __declspec(dllexport)
  #else
	 #define BPPLIB_DLLEXP __declspec(dllimport)
  #endif
  #define BPPAPI WINAPI
#else // no WIN32
	#define BPPLIB_DLLEXP
	#define BPPAPI
#endif 
#ifndef LLONG_MAX 
	 // Use int64 emulation if needed
	#ifndef EMULATE_I64
		#ifndef  WINCE
				#define EMULATE_I64
		#endif
	#endif
#endif
#include <bpptchar.h>

#ifndef _UNICODE
	typedef char BppChar;
	typedef unsigned char BppUChar;
#else
	typedef wchar_t BppChar;
	typedef unsigned short BppUChar;
#endif



#ifdef __TURBOC__

#if (__TURBOC__ < 0x0450)
#ifndef __OLDCPP__
#define __OLDCPP__
#endif
#endif

  enum bool
  {
	 false = 0,
	 true = 1
  };

#endif
#ifdef __OLDCPP__

#define NOCPPEXCEPTION

#else
  // Check for 64 bit Platform
	#if ((SIZE_MAX > UINT_MAX) && (USHRT_MAX < UINT_MAX))
	#define BPP_I64
	#endif
#endif
// Define build specific suffix
#ifndef __PLFEXT
#ifdef BPP_I64
#define __PLFEXT _T("_64")
#else
#define __PLFEXT _T("")
#endif
#endif
#ifndef BPP_DLLWILD
#ifdef _UNICODE
  #ifdef _DEBUG
	 #define BPP_DLLWILD (L"ud" __PLFEXT)
  #else
	 #define BPP_DLLWILD (L"u" __PLFEXT)
  #endif
#else
  #ifdef _DEBUG
	 #define BPP_DLLWILD ("d" __PLFEXT)
  #else
	 #define BPP_DLLWILD __PLFEXT
  #endif
#endif
#endif
	typedef unsigned char BppByte;
	typedef char BppSByte;
	typedef unsigned short BppWord;
	typedef short BppShort;
	typedef unsigned short BppUShort;
#if (!defined(GCC)) && (INT_MAX == SHRT_MAX)
  typedef long BppInt;
  typedef unsigned long BppUInt;
  typedef unsigned BppNWord;
  typedef BppInt BppIntPtr;
#else
  typedef int BppInt;
  typedef unsigned int BppUInt;
  typedef BppUInt BppNWord;
#endif
  typedef BppUInt BppDWord;
#ifdef GCC
  #define __int64 long long
  #define BppTime_T BppLong
#else
  #define BppTime_T time_t
#endif
#ifdef EMULATE_I64
	 #include <int64.h> 
	typedef Int64 BppLong;
	typedef UInt64 BppULong;
#else
  typedef __int64 BppLong;
  typedef unsigned __int64 BppULong;
  typedef __int64 _Int64;
  typedef unsigned __int64 _UInt64;
#endif
#ifdef GCC
	#ifdef BPP_I64
	  typedef __int64 BppIntPtr
	#else
	 typedef int BppIntPtr;
	#endif
	#include <stdint.h>
	template<class T> BppIntPtr INTPTRCAST(T v) { return BppIntPtr(v); }
	template<class T> BppIntPtr INTPTRCAST(T* v) { return BppIntPtr((uintptr_t)v); }
	//#define INTPTRCAST(p) ((BppIntPtr)(uintptr_t)(p))
#else
  typedef ptrdiff_t BppIntPtr;
	#define INTPTRCAST(p) ((BppIntPtr)(p))
#endif

class BppValue;
class ValPtr;
class BppVirtualMachine;
class BppObject;

enum BP_ERR_CODES
{
	BP_ERR_NOMEM = 1
};

enum _BppObjectNotifyType
{
	BPP_OBJ_NONE = 0,
	BPP_OBJ_CREATE,
	BPP_OBJ_COPY,
	BPP_OBJ_DELETE,
	BPP_OBJ_BEFORESERIALIZE,
	BPP_OBJ_AFTERDESERIALIZE
};
typedef enum _BppObjectNotifyType BppObjectNotifyType;

#ifdef NOASM
typedef BppIntPtr (BPPAPI *DLLFUNC0)();
typedef BppIntPtr (BPPAPI *DLLFUNC1)(BppIntPtr);
typedef BppIntPtr (BPPAPI *DLLFUNC2)(BppIntPtr,BppIntPtr);
typedef BppIntPtr (BPPAPI *DLLFUNC3)(BppIntPtr,BppIntPtr,BppIntPtr);
typedef BppIntPtr (BPPAPI *DLLFUNC4)(BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr);
typedef BppIntPtr (BPPAPI *DLLFUNC5)(BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr);
typedef BppIntPtr (BPPAPI *DLLFUNC6)(BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr);
typedef BppIntPtr (BPPAPI *DLLFUNC7)(BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr);
typedef BppIntPtr (BPPAPI *DLLFUNC8)(BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr);
typedef BppIntPtr (BPPAPI *DLLFUNC9)(BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr);
typedef BppIntPtr (BPPAPI *DLLFUNC10)(BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr);
typedef BppIntPtr (BPPAPI *DLLFUNC11)(BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr);
typedef BppIntPtr (BPPAPI *DLLFUNC12)(BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr);
typedef BppIntPtr (BPPAPI *DLLFUNC13)(BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr);
typedef BppIntPtr (BPPAPI *DLLFUNC14)(BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr);
typedef BppIntPtr (BPPAPI *DLLFUNC15)(BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr);
typedef BppIntPtr (BPPAPI *DLLFUNC16)(BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr,BppIntPtr);
#endif

typedef bool (*BPPUSERFUNC) (BppVirtualMachine* vm, int argc, ValPtr& sp);
#if (defined(GCC) && __cplusplus > 0) || defined(__cplusplus)
#if defined(__cpp_rvalue_references) || _MSC_VER >= 1600 || __cplusplus >= 201103
  #define CPP11
  #include <utility>
  #define BPPMOVE(V) std::move((V))
  #define BPPNOEXCEPT noexcept
#else
	#define BPPMOVE(V) (V)
	#define BPPNOEXCEPT
#endif
typedef void (BPPAPI * BppObjNotifyProc)(BppVirtualMachine* vm, BppObject* obj, void** nativeObject, BppObjectNotifyType reason);
typedef bool (BPPAPI *BppDbgWaitProc)(BppVirtualMachine* vm);
typedef bool (BPPAPI *BppBreakRequestProc)(BppVirtualMachine* vm);
typedef bool (BPPAPI *BppLineChangedProc)(BppVirtualMachine* vm, const char* srcName, BppInt lineNumber);
typedef void (BPPAPI *BppRegisterNativeModuleFunc)(BppVirtualMachine* vm);
typedef void (BPPAPI *BppPrinterCallback) (const char* text);
typedef int (BPPAPI *BppCharInputCallback) ();
typedef const char*  (BPPAPI *BppStringInputCallback) ();
#include <locale.h>
struct _CLocale {
#ifdef HAVE_THREADLOCALE
	int _oldmode;
#endif
		char* _old;
		_CLocale() {
#ifdef HAVE_THREADLOCALE
			_oldmode = _configthreadlocale(_ENABLE_PER_THREAD_LOCALE);
#endif
			_old = setlocale(LC_ALL, "C"); 
		} 
		~_CLocale() { 
			setlocale(LC_ALL, _old);
#ifdef HAVE_THREADLOCALE
		  if (_oldmode > 0) _configthreadlocale(_oldmode);
#endif
		} 
};
#define C_LOCALE _CLocale __cLocale

#else
typedef void (BPPAPI *BppObjNotifyProc)void* pVm, void* pObj, void** nativeObject, BppObjectNotifyType reason);
typedef int (BPPAPI *BppDbgWaitProc)();
typedef int (BPPAPI *BppBreakRequestProc)();
typedef int (BPPAPI *BppLineChangedProc)(char* srcName, BppInt lineNumber);
typedef void (BPPAPI *BppRegisterNativeModuleFunc)(void* pVm);
typedef void (BPPAPI *BppPrinterCallback) (char* text);
#endif
#if defined(UNIX)
typedef void* HDLL;
typedef void* HMODULE;
typedef int BOOL;
typedef BppIntPtr (*FARPROC)();
#include <dlfcn.h>
#define LoadLibrary(lpFileName) dlopen(lpFileName, RTLD_NOW)
#define FreeLibrary(hModule) dlclose(hModule)
#define GetProcAddress(hModule,lpProcName) dlsym(hModule,lpProcName)
#endif
#endif // BPPDEF_H
