#ifndef BPPSCAN_H
#define BPPSCAN_H

#include <bppbuf.h>
#include <bppstrng.h>
#include <bppvalue.h>
#include <stdio.h>
enum BppToken
{
 _TMIN  = 0,
 T_NOTOKEN = _TMIN,
 T_ERROR,
 T_EOF,
 T_STRING,
 T_IDENTIFIER,
 T_NUMBER,
 T_FNUMBER,
 T_CLASS,
 T_STATIC,
 T_IF,
 T_ELSE,
 T_WHILE,
 T_RETURN,
 T_FOR,
 T_BREAK,
 T_CONTINUE,
 T_DO,
 T_SWITCH,
 T_CASE,
 T_DEFAULT,
 T_NEW,
 T_DELETE,
 T_TRY,
 T_CATCH,
 T_THROW,
 T_VAR,
 T_FUNCTION,
 T_OPERATOR,
 T_EVAL,
 T_PUBLIC,
 T_PROTECTED,
 T_PRIVATE,
 T_TRON,
 T_TROFF,
 T_TRSTEP,
 T_NIL,
 T_VOID,
 T_AUTO,
/* T_REMARK, */
 T_LE,	/* '<=' */
 T_EQ,	/* '==' */
 T_NE,	/* '!=' */
 T_GE,	/* '>=' */
 T_SHL,	/* '<<' */
 T_SHR,	/* '>>' */
 T_AND,	/* '&&' */
 T_OR,	/* '||' */
 T_INC,	/* '++' */
 T_DEC,	/* '--' */
 T_ADDEQ,	/* '+=' */
 T_SUBEQ,	/* '-=' */
 T_MULEQ,	/* '*=' */
 T_DIVEQ,	/* '/=' */
 T_REMEQ , // '%='
 T_BANDEQ, // &=
 T_BOREQ, // |=
 T_XOREQ, // ^=
 T_BNOTEQ, // ~=
 T_SHLEQ, // <<=
 T_SHREQ, // >>=
 //T_REFEQ, // =>
 //T_UNREFEQ, // =*
 T_CC,	/* '::' */
 T_MEMREF,	/* '->' */
 T_ELLIPSE, // ...
 T_ANY,  // any other character
 _TMAX	= T_ANY
};


class BppCharReader;
class BppCompiler;
struct BppSavedToken;
//enum BppValueType;
class BppValue;
class BPPLIB_DLLEXP BppScanner
{
	friend class BppCompiler;
public:
	BppScanner(FILE* fp, const BppChar* srcName = _T(""), long lnum = 0, unsigned lpos = 0);
	BppScanner(const BppChar* src, const BppChar* srcName = _T(""), long lnum = 0, unsigned lpos = 0);
	~BppScanner();
	BppToken getToken(bool ignoreBuffer = false);
	const BppChar* getTokenName(BppToken tkn) const;
	const BppChar* getTokenString() const;
	double getDouble() const;
	BppInt getLong() const;
	BppUInt getULong() const;
	BppLong getLLong() const;
	BppULong getULLong() const;
	BppValue getNumberValue() const;
	BppChar getChar() const;
	BppToken parseError(const BppChar* msg);
	bool back(bool append = false);
	void push(FILE* fp, const BppChar* srcName = _T(""));
	void push(const BppChar* src, const BppChar* srcName = _T(""));
	const BppChar* getSourceName() const;
	long getLineNumber() const;
	unsigned getLinePos() const;
	BppValueType getNumberType() const;
private:
	BppToken readToken();
	BppChar get();
	BppChar skipSpaces();
	BppToken readString();
	BppToken readChar();
	BppChar readLiteralChar();
	BppToken readIdentifier(BppChar firstChar);
	BppToken readNumber(BppChar firstChar, bool isCharcode = false);
	static bool isIdChar(BppUChar ch);
	BppCharBuffer _stringBuffer;
	double _dblBuffer;
	BppValueType _numType; 
	union {
		BppInt   _intBuffer;
	  BppUInt  _uintBuffer;
	  _Int64   _longBuffer;
	  _UInt64  _ulongBuffer;
	};
	BppChar _charBuffer[2];
	BppCharReader* _reader;
	BppSavedToken* _tokenStack; // stack of look ahead tokens
	BppToken _token;        /* current token */
	//bool     _getBack;     // get last token again, set by back(), reset by getToken() 
	BppChar _savedCh;      /* look ahead character */
	BppChar _lastCh;       /* last input character */
	BppCharBuffer _line;      /* last input line */
	unsigned _linepos;	   /* index in line */
	long _lnum;	           /* line number */
	BppScanner* _next;      // pointer to next (embedded) scanner
	BppString _srcName;    // source name (e.g. filename)
	BppString _funcName;   // name of currently compiled function
	BppString _callerDir;  // working directory when scanner is created
	static BppCompiler* _notifyCompiler;
#ifndef DOXY
	BppScanner(const BppScanner&);
	BppScanner& operator=(const BppScanner&);
#endif
};

#endif // BPPSCAN_H
