#include "bppcomp.h"
#include <limits.h>
#include <stdlib.h>
#include <math.h>
#include "bppvec.h"
#include "bppclass.h"
#include "bppscan.h"
#include "bppvalue.h"
#include "bppbuf.h"
#include "bppufunc.h"
#include "bppprint.h"
#include "dictio.h"
#include "bppvm.h"
#include "bppdbg.h"
#include "bppconst.h"


// literal ::= "null" | "nil" | <string> | <intnumber> | <floatnumber> | <arrayliteral> | <dictionaryliteral> | 
//              <funcliteral> | <literalexpr> | <litid> .
// arrayliteral ::= "[" [ <literal> {"," <literal> } ] "]" .
// dictionaryliteral ::= "{" [ <keyident> ":" <literal> {"," <keyident> ":" <literal> } ] "}" .
// keyident ::= <identifier> | <string> .
// funcliteral ::= "function" <functionbody> .
// literalexpr ::= "#" "(" <litexpr> ")" .
// litid ::= <identifier> . // previously defined named literal value
bool BppCompiler::_doLiteralValue(BppToken token, BppDictionaryEntry& entry, int& errCnt)
{
	int savedErrCnt = errCnt;
	switch (token)
	{
	case T_NIL:
		entry.setValue(BppValue());
		break;
	case T_NUMBER:
	case T_FNUMBER:
		entry.setValue(_scanner->getNumberValue());
		break;
	case T_STRING:
		entry.setValue(new BppString(_scanner->getTokenString()));
		break;
	case T_FUNCTION:
		{
		  BppClass* pclass = 0;
		  token = _scanner->getToken();
		  if ((token == T_ANY) && (_scanner->getChar() == _T('[')))
		  {
			  token = _scanner->getToken();
			  if ((token != T_STRING) && (!_require(T_IDENTIFIER,token)))
			  {
				  errCnt++;
				  break;
			  }
			  pclass = _findClass(_scanner->getTokenString(),errCnt);
			 if (!_fetchRequire(T_ANY,_T(']')))
			 {
				errCnt++;
				break;
			 }

		  }
		  else _scanner->back();
		  BppValueType rtype = BVT_NULL;
		  BppClass* prClass = 0;
		  token = _scanner->getToken();
		  switch (token)
		  {
		  case T_VAR:
		  case T_NIL:
			  break;
		  case T_VOID:
		  case T_FUNCTION:
		  case T_IDENTIFIER:
			  rtype = _vm.findVTypeFromName(_scanner->getTokenString(),true,&prClass);
			  if (rtype == BVT_NULL)
			  {
				  _error(BppString::format(_T("invalid typename %s"),_scanner->getTokenString()).c_str(), &errCnt);
			  }
			  break;
		  default:
			  _scanner->back();
		  }

			if (!_fetchRequire(T_ANY,_T('(')))
			{
				errCnt++;
			}
			else
			{
			 BppVector* codevec = _codeFunctionBody(_T("<anonymous>"),errCnt,pclass,rtype,prClass);
			 if (codevec != 0)
			 {
				 entry.setValue(BppValue(codevec,true));
			 }
		 }
		}
		break;
	case T_ANY:
		{ 
			if (_scanner->getChar()==_T('[')) // array literal
			{
				BppVector* pvec = new BppVector();
				entry.setValue(pvec);
				BppDictionaryEntry dummyEntry(0,BppValue());
				for (token = _scanner->getToken();
					!((token == T_ANY) && (_scanner->getChar() == _T(']')));
					token = _scanner->getToken()) 
				{
					// read single literal value
					_doLitExpr(token,dummyEntry,errCnt);
					pvec->push_back(dummyEntry.getValue());
					if (errCnt > savedErrCnt)
						break;
					// value must be followed by , or ]
					token = _scanner->getToken();
					if (token != T_ANY) // not a single character
					{
						_error(_T("',' or ']' expected"), &errCnt);
						break;
					}
					if (_scanner->getChar() == _T(']')) // return if end is reached
						break;
					if (!_require(T_ANY,token,_T(','))) // return if comma missing
					{
						errCnt++;
						break;
					}
				} // end for
				break;
			} // end array literal
			else if (_scanner->getChar()==_T('{')) // dictionary literal
			{
				BppDictionary* pdict = new BppDictionary();
				entry.setValue(pdict);
				for (token = _scanner->getToken();
					!((token == T_ANY) && (_scanner->getChar() == _T('}')));
					token = _scanner->getToken()) 
				{
					if (!((token == T_IDENTIFIER) || _require(T_STRING,token)))
					{ 
						errCnt++;
						break;
					}
					BppDictionaryEntry& entryref = (*pdict)[_scanner->getTokenString()];
					if (!_fetchRequire(T_ANY,_T(':')))
					{ 
						errCnt++;
						break;
					}
					_doLitExpr(_scanner->getToken(),entryref,errCnt);
					if (errCnt > savedErrCnt)
						break;
					// value must be followed by , or }
					token = _scanner->getToken();
					if (token != T_ANY) // not a single character
					{
						_error(_T("',' or '}' expected"), &errCnt);
						break;
					}
					if (_scanner->getChar() == _T('}')) // return if end is reached
						break;
					if (!_require(T_ANY,token,_T(','))) // return if comma missing
					{
						errCnt++;
						break;
					}
				} // end for
				break;
			} // end dictionary literal
			else if (_scanner->getChar()==_T('#')) // literal expression
			{
				if (!_fetchRequire(T_ANY,_T('(')))
				{
					errCnt++;
					break;
				}
				token = _scanner->getToken();
				if (!_doLitExpr(token,entry,errCnt)) return false;
				if (!_fetchRequire(T_ANY,_T(')'))) errCnt++;
				break;
			} // end literal expression
			_require(T_IDENTIFIER, token);
			errCnt++;
			break;
		} // end T_ANY
	case T_IDENTIFIER:
		{
			const BppChar* str = _scanner->getTokenString();
			if (_tcscmp(str, _T("defined")) == 0)
			{
				token = _scanner->getToken();
				bool bracket = false;
				if ((token == T_ANY) && (_scanner->getChar() == _T('('))) {
					bracket = true;
				  token =  _scanner->getToken();
				}
				if (!_require(T_IDENTIFIER, token)) errCnt++;
				BppDictionaryEntry* lit = _namedLiterals.find(_scanner->getTokenString());
				entry.setValue(BppInt(lit ? 1 : 0));
				if (bracket && !_fetchRequire(T_ANY, _T(')'))) errCnt++;
				break;
			}
			BppDictionaryEntry* lit = _namedLiterals.find(str);
			if (lit != 0)
			{
				entry.setValue(lit->getValue());
				break;
			}
		}
	default:
		_error(_T("literal expected"), &errCnt);
	} // end switch
	return true;
}


/// litexpr ::= litexpr1 { "?" litexpr ":" litexpr } .
bool BppCompiler::_doLitExpr(BppToken token, BppDictionaryEntry& entry, int& errCnt)
{
	if (!_doLitExpr1(token,entry,errCnt)) return false;
	while (((token = _scanner->getToken()) == T_ANY) && (_scanner->getChar()==_T('?')))
	{
		BppDictionaryEntry de1(0,BppValue()), de2(0,BppValue());
		if (!_doLitExpr(_scanner->getToken(),de1,errCnt)) return false;
		if (!_fetchRequire(T_ANY,_T(':')))
		{
			++errCnt;
			break;
		}
		if (!_doLitExpr(_scanner->getToken(),de2,errCnt)) return false;
		if (((BppInt)entry.getValue()) != 0) entry.setValue(de1.getValue());
		else entry.setValue(de2.getValue());
	}
	_scanner->back();
	return true;
}

/// litexpr1 ::= litexpr2 { "||" litexpr2 } .
bool BppCompiler::_doLitExpr1(BppToken token, BppDictionaryEntry& entry, int& errCnt)
{
	if (!_doLitExpr2(token,entry,errCnt)) return false;
	while (_scanner->getToken() == T_OR)
	{
		BppDictionaryEntry de(0,BppValue());
		if (!_doLitExpr2(_scanner->getToken(),de,errCnt)) return false;
		if (((BppInt) entry.getValue()) == 0) entry.setValue(de.getValue());
	}
	_scanner->back();
	return true;
}

/// litexpr2 ::= litexpr3 { "&&" litexpr3 } .
bool BppCompiler::_doLitExpr2(BppToken token, BppDictionaryEntry& entry, int& errCnt)
{
	if (!_doLitExpr3(token,entry,errCnt)) return false;
	while (_scanner->getToken() == T_AND)
	{
		BppDictionaryEntry de(0,BppValue());
		if (!_doLitExpr3(_scanner->getToken(),de,errCnt)) return false;
		if (((BppInt) entry.getValue()) != 0) entry.setValue(de.getValue());
	}
	_scanner->back();
	return true;
}

/// litexpr3 ::= litexpr4 { "|" litexpr4 } .
bool BppCompiler::_doLitExpr3(BppToken token, BppDictionaryEntry& entry, int& errCnt)
{
	if (!_doLitExpr4(token,entry,errCnt)) return false;
	while (((token = _scanner->getToken()) == T_ANY) && (_scanner->getChar()==_T('|')))
	{
		BppDictionaryEntry de(0,BppValue());
		if (!_doLitExpr4(_scanner->getToken(),de,errCnt)) return false;
		entry.setValue( ((BppInt) entry.getValue()) | ((BppInt) de.getValue()) );
	}
	_scanner->back();
	return true;
}

/// litexpr4 ::= litexpr5 { "^" litexpr5 } .
bool BppCompiler::_doLitExpr4(BppToken token, BppDictionaryEntry& entry, int& errCnt)
{
	if (!_doLitExpr5(token,entry,errCnt)) return false;
	while (((token = _scanner->getToken()) == T_ANY) && (_scanner->getChar()==_T('^')))
	{
		BppDictionaryEntry de(0,BppValue());
		if (!_doLitExpr5(_scanner->getToken(),de,errCnt)) return false;
		entry.setValue( ((BppInt) entry.getValue()) ^ ((BppInt) de.getValue()) );
	}
	_scanner->back();
	return true;
}

/// litexpr5 ::= litexpr6 { "&" litexpr6 } 
bool BppCompiler::_doLitExpr5(BppToken token, BppDictionaryEntry& entry, int& errCnt)
{
	if (!_doLitExpr6(token,entry,errCnt)) return false;
	while (((token = _scanner->getToken()) == T_ANY) && (_scanner->getChar()==_T('&')))
	{
		BppDictionaryEntry de(0,BppValue());
		if (!_doLitExpr6(_scanner->getToken(),de,errCnt)) return false;
		entry.setValue( ((BppInt) entry.getValue()) & ((BppInt) de.getValue()) );
	}
	_scanner->back();
	return true;
}

static BppValueType getCommonLitExprType(BppValueType tl, BppValueType tr)
{
	if (tl == BVT_NULL || tr == BVT_NULL) return BVT_NULL;
	if ((tl == BVT_STRING || tl == BVT_STRINGREF)
		&& (tr == BVT_STRING || tr == BVT_STRINGREF)) return BVT_STRING;
	BppValueType rt = tl < tr ? tr : tl;
	if (rt == BVT_STRING || rt == BVT_STRINGREF) return BVT_STRING;
#ifdef BPP_I64
	if (rt == BVT_INTPTR)
	{
		if (tl == BVT_ULONG || tr == BVT_ULONG) rt = BVT_ULONG;
	}
	else 
#endif
		if (rt < BVT_INT) rt = BVT_INT;
	if (rt > _BVTVALMAX) rt = BVT_INTPTR;
	return rt;
}

#define cmpres(T) T(vl) < T(vr) ? -1 : T(vr) < T(vl) ? -1 : 0
static BppInt cmpLiteralValues(const BppValue& vl, const BppValue& vr)
{
	BppValueType tl = vl.getType();
	BppValueType tr = vr.getType();
	if (tl == BVT_NULL) return (tr == BVT_NULL) ? 0 : -1;
	if (tr == BVT_NULL) return (tl == BVT_NULL) ? 0 : 1;
	BppValueType rt = getCommonLitExprType(tl,tr);
	BppInt result;
	switch (rt)
	{
	case BVT_INT:
		result = cmpres(BppInt);
		break;
	case BVT_UINT:
		result = cmpres(BppUInt);
		break;
	case BVT_LONG:
		result = cmpres(BppLong);
		break;
	case BVT_ULONG:
		result = cmpres(BppULong);
		break;
	case BVT_INTPTR:
#ifdef GCC
		result = cmpres(BppLong);
#else
		result = cmpres(BppIntPtr);
#endif   
		break;
	case BVT_FLOAT:
	case BVT_DOUBLE:
		result = cmpres(double);
		break;
	case BVT_STRING:
		{
			BppString* sl = vl;
			BppString* sr = vr;
			if (sl && sr)
			{
			 result = (BppInt) (sl->compare(*sr));
			 break;
			}
		}
	default:
#ifdef GCC
		result = cmpres(BppLong);
#else
		result = cmpres(BppIntPtr);
#endif   
			break;
	}
	return result;
}
#undef cmpres

/// litexpr6 ::= litexpr7 { ("==" | "!=") litexpr7 } .
bool BppCompiler::_doLitExpr6(BppToken token, BppDictionaryEntry& entry, int& errCnt)
{
	if (!_doLitExpr7(token,entry,errCnt)) return false;
	while ((token = _scanner->getToken()) == T_EQ || token == T_NE)
	{
		BppDictionaryEntry de(0,BppValue());
		if (!_doLitExpr7(_scanner->getToken(),de,errCnt)) return false;
		BppInt cmp = cmpLiteralValues(entry.getValue(),de.getValue());
		if (token == T_EQ)
			entry.setValue( (BppInt)((cmp == 0) ? 1 : 0 ));
		else
			entry.setValue( (BppInt)((cmp != 0) ? 1 : 0 ));
	}
	_scanner->back();
	return true;
}

/// litexpr7 ::= litexpr8 { ("<" | ">" | "<=" | ">="} litexpr8 } .
bool BppCompiler::_doLitExpr7(BppToken token, BppDictionaryEntry& entry, int& errCnt)
{
	if (!_doLitExpr8(token,entry,errCnt)) return false;
	BppChar c = 0;
	while ((token = _scanner->getToken()) == T_LE 
		|| token == T_GE 
		|| ((token == T_ANY) && (((c = _scanner->getChar()) == _T('>')) || (c == _T('<'))))
		) 
	{
		BppDictionaryEntry de(0,BppValue());
		if (!_doLitExpr8(_scanner->getToken(),de,errCnt)) return false;
		BppInt cmp = cmpLiteralValues(entry.getValue(),de.getValue());
		switch (token)
		{
		case T_LE:
			entry.setValue( (BppInt)((cmp <= 0) ? 1 : 0 ));
			break;
		case T_GE:
			entry.setValue( (BppInt)((cmp >= 0) ? 1 : 0 ));
			break;
		default:
			if (c == _T('<'))
				entry.setValue( (BppInt)((cmp < 0) ? 1 : 0 ));
			else
				entry.setValue( (BppInt)((cmp > 0) ? 1 : 0 ));
		}
	}
	_scanner->back();
	return true;
}

/// litexpr8 ::= litexpr9 { ("<<" | ">>") litexpr9 } .
bool BppCompiler::_doLitExpr8(BppToken token, BppDictionaryEntry& entry, int& errCnt)
{
	if (!_doLitExpr9(token,entry,errCnt)) return false;
	while ((token = _scanner->getToken()) == T_SHL || token == T_SHR) 
	{
		BppDictionaryEntry de(0,BppValue());
		if (!_doLitExpr9(_scanner->getToken(),de,errCnt)) return false;
		if (token == T_SHL)
			entry.setValue( ((BppInt) entry.getValue()) << ((BppInt) de.getValue()) );
		else
			entry.setValue( ((BppInt) entry.getValue()) >> ((BppInt) de.getValue()) );
	}
	_scanner->back();
	return true;
}

static BppValue addLiteralValues(const BppValue& vl, const BppValue& vr)
{
	BppValueType tl = vl.getType();
	BppValueType tr = vr.getType();
	BppValueType rt = getCommonLitExprType(tl,tr);
	switch (rt)
	{
	case BVT_INT:
		return ((BppInt) vl) + ((BppInt) vr);
	case BVT_UINT:
		return ((BppUInt) vl) + ((BppUInt) vr);
	case BVT_LONG:
		return BppLong(vl) + BppLong(vr);
	case BVT_ULONG:
		return BppULong(vl) + BppULong(vr);
	case BVT_INTPTR:
#ifdef GCC
		return BppValue(BppLong(vl) + BppLong(vr),true);
#else
		return BppValue(BppIntPtr(vl) + BppIntPtr(vr),true);
#endif
	case BVT_FLOAT:
		return ((float) vl) + ((float) vr);
	case BVT_DOUBLE:
		return ((double) vl) + ((double) vr);
	case BVT_STRING:
		{
			BppString* sl = vl;
			BppString* sr = vr;
			if (sl)
			{
				BppString* s = new BppString(*sl);
				if (sr)
				{
						 (*s) += (*sr);
				}
				else if (vr.isIntegralType())
				{
					(*s) += (BppChar) (BppInt) vr;
				}
				return s;
			}
			else if (sr && vl.isIntegralType())
			{
					BppChar c[2];
					// left operand becomes first char of new string
					c[0] = (BppChar) (BppInt) vl;;
					c[1] = 0;
					BppString* s = new BppString(c);
					(*s) += sr->c_str();
					return s;
			}
		}
	}
	return vl;
}

static BppValue subLiteralValues(const BppValue& vl, const BppValue& vr)
{
	BppValueType tl = vl.getType();
	BppValueType tr = vr.getType();
	BppValueType rt = getCommonLitExprType(tl,tr);
	switch (rt)
	{
	case BVT_INT:
		return ((BppInt) vl) - ((BppInt) vr);
	case BVT_UINT:
		return ((BppUInt) vl) - ((BppUInt) vr);
	case BVT_LONG:
		return BppLong(vl) - BppLong(vr);
	case BVT_ULONG:
		return BppULong(vl) - BppULong(vr);
	case BVT_INTPTR:
#ifdef GCC
		return BppValue(BppLong(vl) - BppLong(vr),true);
#else
		return BppValue(BppIntPtr(vl) - BppIntPtr(vr),true);
#endif
	case BVT_FLOAT:
		return ((float) vl) - ((float) vr);
	case BVT_DOUBLE:
		return ((double) vl) - ((double) vr);
	}
  return vl;
}

static BppValue mulLiteralValues(const BppValue& vl, const BppValue& vr)
{
	BppValueType tl = vl.getType();
	BppValueType tr = vr.getType();
	BppValueType rt = getCommonLitExprType(tl,tr);
	switch (rt)
	{
	case BVT_INT:
		return ((BppInt) vl) * ((BppInt) vr);
	case BVT_UINT:
		return ((BppUInt) vl) * ((BppUInt) vr);
	case BVT_LONG:
		return BppLong(vl) * BppLong(vr);
	case BVT_ULONG:
		return BppULong(vl) * BppULong(vr);
	case BVT_FLOAT:
		return ((float) vl) * ((float) vr);
	case BVT_DOUBLE:
		return ((double) vl) * ((double) vr);
	}
	return (
 #ifdef GCC
	BppLong
#else  
	BppIntPtr
#endif
	) 0;
}

static BppValue divLiteralValues(const BppValue& vl, const BppValue& vr)
{
	BppValueType tl = vl.getType();
	BppValueType tr = vr.getType();
	BppValueType rt = getCommonLitExprType(tl,tr);
	switch (rt)
	{
	case BVT_INT:
		return ((BppInt) vl) / ((BppInt) vr);
	case BVT_UINT:
		return ((BppUInt) vl) / ((BppUInt) vr);
	case BVT_LONG:
		return BppLong(vl) / BppLong(vr);
	case BVT_ULONG:
		return BppULong(vl) / BppULong(vr);
	case BVT_FLOAT:
		return ((float) vl) / ((float) vr);
	case BVT_DOUBLE:
		return ((double) vl) / ((double) vr);
	}
	return (
#ifdef GCC
	BppLong
#else
	 BppIntPtr
#endif
	) 0;
}

static BppValue modLiteralValues(const BppValue& vl, const BppValue& vr)
{
	BppValueType tl = vl.getType();
	BppValueType tr = vr.getType();
	BppValueType rt = getCommonLitExprType(tl,tr);
	switch (rt)
	{
	case BVT_INT:
		return ((BppInt) vl) % ((BppInt) vr);
	case BVT_UINT:
		return ((BppUInt) vl) % ((BppUInt) vr);
	case BVT_LONG:
		return BppLong(vl) % BppLong(vr);
	case BVT_ULONG:
		return BppULong(vl) % BppULong(vr);
	case BVT_FLOAT:
		return fmod((float) vl,(float) vr);
	case BVT_DOUBLE:
		return fmod((double) vl, (double) vr);
	}
	return (
#ifdef GCC
	BppLong
#else
	 BppIntPtr
#endif
	) 0;
}

/// litexpr9 ::= litexpr10 { ("+" | "-") litexpr10 } .
bool BppCompiler::_doLitExpr9(BppToken token, BppDictionaryEntry& entry, int& errCnt)
{
	if (!_doLitExpr10(token,entry,errCnt)) return false;
	BppChar c = 0;
	while (((token = _scanner->getToken()) == T_ANY) && (((c = _scanner->getChar()) == _T('+')) || (c == _T('-')))) 
	{
		BppDictionaryEntry de(0,BppValue());
		if (!_doLitExpr10(_scanner->getToken(),de,errCnt)) return false;
		if (c == _T('+'))
			entry.setValue( addLiteralValues(entry.getValue(),de.getValue()) );
		else
			entry.setValue( subLiteralValues(entry.getValue(),de.getValue()) );
	}
	_scanner->back();
	return true;
}

/// litexpr10 ::= litexpr11 { ("*" | "/" | "%") litexpr11 } .
bool BppCompiler::_doLitExpr10(BppToken token, BppDictionaryEntry& entry, int& errCnt)
{
	if (!_doLitExpr11(token,entry,errCnt)) return false;
	BppChar c = 0;;
	while (((token = _scanner->getToken()) == T_ANY) && (((c = _scanner->getChar()) == _T('*')) || (c == _T('/')) || (c == _T('%'))))  
	{
		BppDictionaryEntry de(0,BppValue());
		if (!_doLitExpr11(_scanner->getToken(),de,errCnt)) return false;
		if (c == _T('*'))
			entry.setValue( mulLiteralValues(entry.getValue(),de.getValue()) );
		else if (c == _T('/'))
			entry.setValue( divLiteralValues(entry.getValue(),de.getValue()) );
		else
			entry.setValue( modLiteralValues(entry.getValue(),de.getValue()) );
	}
	_scanner->back();
	return true;
}

/// litexpr11 ::= [ "+" | "-" | "!" | "~" ] litexpr12 .
bool BppCompiler::_doLitExpr11(BppToken token, BppDictionaryEntry& entry, int& errCnt)
{
	if (token == T_ANY)
	{
		BppChar c = _scanner->getChar();
		if ( (c==_T('+')) || (c==_T('-')) || (c==_T('!')) || (c==_T('~')) )
		{
			BppDictionaryEntry de(0,BppValue());
			if (!_doLitExpr11(_scanner->getToken(),de,errCnt)) return false;
			const BppValue& v = de.getValue();
		 BppValueType vt = v.getType();
		 vt = getCommonLitExprType(vt,vt);
			switch (c)
			{
			case _T('+'):
			 switch (vt)
			 {
			 case BVT_UINT:
				 entry.setValue((BppUInt) v);
				 break;
			 case BVT_LONG:
				 entry.setValue((BppLong) v);
				 break;
			 case BVT_ULONG:
				 entry.setValue((BppULong) v);
				 break;
			 case BVT_INTPTR:
				 entry.setValue((
#ifdef GCC
					BppLong
#else
					BppIntPtr
#endif
					) v);
				 break;
			 case BVT_FLOAT:
				 entry.setValue((float) v);
				 break;
			 case BVT_DOUBLE:
				 entry.setValue((double) v);
				 break;
			 default:
				 entry.setValue((BppInt) v);
			 }
			 break;
			case _T('-'):
			 switch (v.getType())
			 {
			 case BVT_UINT:
				 entry.setValue(-((BppUInt) v));
				 break;
			 case BVT_LONG:
				 entry.setValue(-((BppLong) v));
				 break;
			 case BVT_ULONG:
				 entry.setValue(-((BppULong) v));
				 break;
			 case BVT_INTPTR:
#ifdef GCC
				 entry.setValue(-((BppLong) v));
#else
				 entry.setValue(-((BppIntPtr) v));
#endif 
				 break;
			 case BVT_FLOAT:
				 entry.setValue(- ((float) v));
				 break;
			 case BVT_DOUBLE:
				 entry.setValue(- ((double) v));
				 break;
			 default:
				 entry.setValue(- ((BppInt) v));
			 }
				break;
			case _T('!'):
				entry.setValue(((BppInt) v) ? (BppInt) 0 : (BppInt) 1);
				break;
			case _T('~'):
			 switch (v.getType())
			 {
			 case BVT_UINT:
				 entry.setValue(~((BppUInt) v));
				 break;
			 case BVT_LONG:
				 entry.setValue(~((BppLong) v));
				 break;
			 case BVT_ULONG:
				 entry.setValue(~((BppULong) v));
				 break;
			 case BVT_INTPTR:
#ifdef GCC
				 entry.setValue(-((BppLong) v));
#else
				 entry.setValue(-((BppIntPtr) v));
#endif 
			 default:
				 entry.setValue(~ ((BppInt) v));
			 }
				break;
			}
			return true;
		}
	}
	return _doLitExpr12(token,entry,errCnt);
}

/// litexpr12 ::= ( "(" litexpr ")" ) | literal 
bool BppCompiler::_doLitExpr12(BppToken token, BppDictionaryEntry& entry, int& errCnt)
{
	BppDictionaryEntry de(0,BppValue());
	if ((token == T_ANY) && (_scanner->getChar() == _T('(')))
	{
		if (!_doLitExpr(_scanner->getToken(),de,errCnt)) return false;
		if (!_fetchRequire(T_ANY,_T(')'))) ++errCnt;
	}
	else if (!_doLiteralValue(token,de,errCnt)) return false;
	entry.setValue(de.getValue());
	return true;
}

