
/** @file
Defines implementation functions for B++ predefined functions.
*/

#ifndef DOXY

#ifndef WINCE
#ifndef __linux__
#include <conio.h>
#endif
#include <time.h>
#endif
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>

#include "bppufunc.h"
#include "bppdfunc.h"
#include "bppvm.h"
#include "bppint.h"
#include "bppvalue.h"
#include "bppbuf.h"
#include "bppclass.h"
#include "bppdict.h"
#include "bppobjct.h"
#include "bppstrng.h"
#include "bppvec.h"
#include "bppprint.h"
#include "bppcomp.h"
#include "dictio.h"
#include "bppfimpl.h"

#ifndef WINCE
#ifdef __MSDOS__
	#include <dos.h>
	#ifdef __TURBOC__
		#include <alloc.h>
	#else
		#include <malloc.h>
	#endif

#endif // __MSDOS__
#endif
#ifdef WIN32
	#include <windows.h>
#endif

static void _registerSymbol(BppVirtualMachine* vm,const BppChar* name,BppSymbolType stype,const BppValue& val)
{
	BppDictionaryEntry& de = vm->symbols()[name]; de.setValue(val); de.setSymbolType(stype);
}
#define ADDFILE(NAME) _registerSymbol(vm,_T(#NAME),ST_SDATA,(FILE*) NAME)
#define ADDFUNC(NAME) _registerSymbol(vm,_T(#NAME),ST_SFUNCTION,(BPPUSERFUNC) &Bpp_##NAME)

void BppDefaultFunctions::registerFiles(BppVirtualMachine* vm)
{
	ADDFILE(stdin);
	ADDFILE(stdout);
	ADDFILE(stderr);
}

void BppDefaultFunctions::registerSysVariables(BppVirtualMachine* vm)
{
	static const BppChar* OSTYPE=
#ifdef WIN32
#ifndef WINCE
		_T("WIN32")
#else
		_T("WINCE")
#endif
#else
#ifdef __MSDOS__
		_T("DOS")
#else
#ifdef LINUX
		_T("LINUX")
#else
#ifdef UNIX
		_T("UNIX")
#else
		_T("")
#endif
#endif
#endif
#endif
		;
	_registerSymbol(vm,_T("__OSTYPE__"),ST_SDATA,new BppString(OSTYPE));
	_registerSymbol(vm,_T("__UNICODE__"),ST_SDATA,(BppInt)
#ifdef _UNICODE
			1
#else
			0
#endif
			);

}

void BppDefaultFunctions::registerDictFunctions(BppVirtualMachine* vm)
{
	ADDFUNC(newdict);
	ADDFUNC(dictcontains);
	ADDFUNC(dicterase);
	ADDFUNC(dictclear);
	ADDFUNC(dictgetkeys);
	ADDFUNC(dictgetvalues);
}

void BppDefaultFunctions::registerMathFunctions(BppVirtualMachine* vm)
{
	ADDFUNC(sin);
	ADDFUNC(cos);
	ADDFUNC(tan);
	ADDFUNC(atan);
	ADDFUNC(sqrt);
	ADDFUNC(log);
	ADDFUNC(exp);
	ADDFUNC(abs);
	ADDFUNC(pow);
}

void BppDefaultFunctions::registerIOFunctions(BppVirtualMachine* vm)
{
	ADDFUNC(fopen);
	ADDFUNC(fclose);
	ADDFUNC(feof);
	ADDFUNC(getc);
	ADDFUNC(putc);
	ADDFUNC(gets);
	ADDFUNC(fgets);
	ADDFUNC(fputs);
	ADDFUNC(print);
	ADDFUNC(freadval);
	ADDFUNC(fwriteval);
	ADDFUNC(fread);
	ADDFUNC(fwrite);
#ifdef WIN32
	ADDFUNC(getwc);
	ADDFUNC(putwc);
	ADDFUNC(fgetws);
	ADDFUNC(fputws);
#endif
}

void BppDefaultFunctions::registerFunctions(BppVirtualMachine* vm)
{
	registerFiles(vm);
	registerSysVariables(vm);
	ADDFUNC(newvector);
	ADDFUNC(newstring);
	ADDFUNC(newbuffer);
	ADDFUNC(newcbuffer);
	ADDFUNC(clone);

	// I/O functions
	registerIOFunctions(vm);

	// string functions
	ADDFUNC(strlen);
	ADDFUNC(strcmp);
	ADDFUNC(stricmp);
	ADDFUNC(strsplit);
	ADDFUNC(strfind);
	ADDFUNC(strerase);
	ADDFUNC(strinsert);
	ADDFUNC(strreplace);
	ADDFUNC(substr);
	ADDFUNC(strupr);
	ADDFUNC(strlwr);

	// vector and buffer related functions
	ADDFUNC(vecsize);
	ADDFUNC(vinsert);
	ADDFUNC(verase);
	ADDFUNC(vswap);
	ADDFUNC(vmove);
	ADDFUNC(vslice);
	ADDFUNC(vclear);
	ADDFUNC(size);
	ADDFUNC(shrink);
	ADDFUNC(Vec);
	ADDFUNC(T);
	ADDFUNC(CBuf);
	ADDFUNC(qsort);
	ADDFUNC(foreach);
	ADDFUNC(find);
	ADDFUNC(rfind);

	// program related functions
	ADDFUNC(compile);
	ADDFUNC(compilefn);
	ADDFUNC(litexpr);
	ADDFUNC(loadmodule);

	// basic type conversion
	ADDFUNC(bool);
	ADDFUNC(sbyte);
	ADDFUNC(byte);
	ADDFUNC(short);
	ADDFUNC(ushort);
	ADDFUNC(int);
	ADDFUNC(uint);
	ADDFUNC(long);
	ADDFUNC(ulong);
	ADDFUNC(intptr);
	ADDFUNC(float);
	ADDFUNC(double);
	ADDFUNC(string);
	ADDFUNC(buffer);
	ADDFUNC(charbuffer);
	ADDFUNC(vector);
	ADDFUNC(dictionary);

	// RTTI
	ADDFUNC(gettypename);
	ADDFUNC(gettype);
	ADDFUNC(getclassname);
	ADDFUNC(dynamic_cast);
	ADDFUNC(getsymbol);

	// date and time
	ADDFUNC(time);
	ADDFUNC(timer);
	// math functions
	registerMathFunctions(vm);

	// miscellanous funcs
	ADDFUNC(rand);
	ADDFUNC(srand);
	ADDFUNC(ctime);
	ADDFUNC(localtime);
	ADDFUNC(getargs);
	ADDFUNC(getusrargs);
	ADDFUNC(argcnt);
	ADDFUNC(arg);
	ADDFUNC(argv);
	ADDFUNC(call);

	// system functions
	ADDFUNC(addr);
	ADDFUNC(getenv);
	ADDFUNC(setenv);
	ADDFUNC(LoadLibrary);
	ADDFUNC(FreeLibrary);
	ADDFUNC(GetProcAddress);
	ADDFUNC(CallLibFunc);
	ADDFUNC(errno);
	ADDFUNC(nerr);
	ADDFUNC(strerror);
	ADDFUNC(setlocale);
	ADDFUNC(memmove);
#ifndef WINCE
	ADDFUNC(system);
#endif
#ifdef WIN32
	ADDFUNC(GetUser32Func);
	ADDFUNC(GetKernel32Func);
	ADDFUNC(GetGdi32Func);
	ADDFUNC(GetShell32Func);
	ADDFUNC(GetOle32Func);
	ADDFUNC(GetLastError);
	ADDFUNC(GetSysErrorMsg);
	vm->registerNativeFunction(_T("GetCoreFunc") ,(BPPUSERFUNC) &Bpp_GetUser32Func);
#endif
	registerDictFunctions(vm);
	// graphic bios functions
	ADDFUNC(setscrmode);
	ADDFUNC(setpixel);
	ADDFUNC(line);
#ifdef __MSDOS__
	registerDosFunctions(vm);
#endif
}

#ifdef __MSDOS__

void BppDefaultFunctions::registerDosFunctions(BppVirtualMachine* vm)
{
	ADDFUNC(memsize);

	ADDFUNC(setreg);
	ADDFUNC(getreg);
	ADDFUNC(segread);
	ADDFUNC(int86);
	ADDFUNC(int86x);
	ADDFUNC(inportb);
	ADDFUNC(outportb);
	ADDFUNC(inport);
	ADDFUNC(outport);

}
#endif
#undef ADDFUNC


// vec = newvector(capacity=0,init=true)
bool BppDefaultFunctions::Bpp_newvector(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	ARGCMAX(2);
	ValPtr bp = sp; bp-=argc;
	bool init = true;
	switch (argc)
	{
	case 0:
		*bp = new BppVector();
		return true;
	case 2:
		init = (((BppInt) bp[2]) != 0);
	default:
		*bp = new BppVector((unsigned) (BppInt) bp[1],init);
		break;
	}
	sp = bp;
	return true;
}
// str = newstring(string|buffer|charbuffer)
// str = newstring(size,fillchar=' ')
// str = newstring(fmtstring, ... )
bool BppDefaultFunctions::Bpp_newstring(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	ARGCMIN(1);
	ValPtr bp = sp; bp -= argc;
	switch (maskWeakRef(bp[1].getType()))
	{
	case BVT_STRING:
		{
			BppString* s = bp[1];
			if (s != 0)
			{
				if (argc > 1)
				{
					ValPtr pv = bp;
					pv += 2;
					*bp = new BppString(s->c_str(),argc-1,pv);
				}
				else
					*bp = new BppString(s->c_str());
			}
			else
				*bp = new BppString();
		}
		break;
	case BVT_BUFFER:
		{
			ARGCMAX(1);
			BppBuffer* b = bp[1];

		 BppString* s = b != 0 ? new BppString(*b) : new BppString();
			*bp = s;
		}
		break;
	case BVT_CHARBUFFER:
		{
			ARGCMAX(1);
			BppCharBuffer* b = bp[1];
		 BppString* s = b != 0 ? new BppString(*b) : new BppString();
			*bp = s;
		}
		break;
	case BVT_INT:
	case BVT_UINT:
	case BVT_SHORT:
	case BVT_USHORT:
	case BVT_SBYTE:
	case BVT_BYTE:
	case BVT_LONG:
	case BVT_ULONG:
	case BVT_INTPTR:
		{
			ARGCMAX(2);
			BppUInt size = bp[1];
			BppChar fill = _T(' ');
			if (argc == 2) fill = (BppChar) (BppUInt) bp[2];
			BppString* s = new BppString();
			*bp = s;
			for (BppUInt i=0;i<size;i++) (*s) += fill;
		}
		break;
	default:
		THROW(_T("invalid value type"));
		break;
	}
	sp = bp;
	return true;
}

// str = newbuffer(string|buffer|charbuffer)
// str = newbuffer(size,fillbyte=0)
bool BppDefaultFunctions::Bpp_newbuffer(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,2);
	ValPtr bp = sp; bp -= argc;
	switch (maskWeakRef(bp[1].getType()))
	{
	case BVT_BUFFER:
		{
			BppBuffer* b = bp[1];
			if (b != 0)
				*bp = new BppBuffer(*b);
			else
				*bp = new BppBuffer();
		}
		break;
	case BVT_STRING:
		{
			BppString* s = bp[1];
			if (s != 0)
			{
				size_t size =  s->length()+1;
				BppBuffer* b = new BppBuffer(size);

				for (size_t i=0;i<size;i++)
					b->push_back((BppByte)(*s)[i]);
				b->push_back(0);
				*bp = b;
			}
			else
				*bp = new BppBuffer();
		}
		break;
	case BVT_CHARBUFFER:
		{
			BppCharBuffer* cb = bp[1];
			if (cb != 0)
			{
				size_t size = cb->size();
				BppBuffer* b = new BppBuffer(size);
				for (size_t i=0;i<size;i++)
					b->push_back((BppByte) (*cb)[i]);
				*bp = b;
			}
			else
				*bp = new BppBuffer();
		}
		break;
	case BVT_INT:
	case BVT_UINT:
	case BVT_SHORT:
	case BVT_USHORT:
	case BVT_SBYTE:
	case BVT_BYTE:
	case BVT_LONG:
	case BVT_ULONG:
	case BVT_INTPTR:
		{
			BppUInt size = bp[1];
			BppByte fill = 0;
			if (argc == 2) fill = (BppByte) bp[2];
			BppBuffer* b = new BppBuffer((unsigned) size);
			*bp = b;
			for (BppUInt i=0;i<size;++i) b->push_back(fill);
		}
		break;
	default:
		THROW(_T("invalid value type"));
		break;
	}
	sp = bp;
	return true;
}

// str = newcbuffer(string|buffer|charbuffer)
// str = newcbuffer(size,fillchar=0)
bool BppDefaultFunctions::Bpp_newcbuffer(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,2);
	ValPtr bp = sp; bp -= argc;
	switch (maskWeakRef(bp->getType()))
	{
	case BVT_CHARBUFFER:
		{
			BppCharBuffer* b = bp[1];
			if (b != 0)
				*bp = new BppCharBuffer(*b);
			else
				*bp = new BppCharBuffer();
		}
		break;
	case BVT_STRING:
		{
			BppString* s = bp[1];
			if (s != 0)
			{
				size_t size =  s->length()+1;
				BppCharBuffer* b = new BppCharBuffer(size);

				for (size_t i=0;i<size;i++)
					b->push_back((*s)[i]);
				b->push_back(0);
				*bp = b;
			}
			else
				*bp = new BppCharBuffer();
		}
		break;
	case BVT_BUFFER:
		{
			BppBuffer* b = bp[1];
			if (b != 0)
			{
				size_t size = b->size();
				BppCharBuffer* cb = new BppCharBuffer(size);
				for (size_t i=0;i<size;i++)
					cb->push_back((BppChar) (*b)[i]);
				*bp = cb;
			}
			else
				*bp = new BppCharBuffer();
		}
		break;
	case BVT_INT:
	case BVT_UINT:
	case BVT_SHORT:
	case BVT_USHORT:
	case BVT_SBYTE:
	case BVT_BYTE:
	case BVT_LONG:
	case BVT_ULONG:
	case BVT_INTPTR:
		{
			BppUInt size = bp[1];
			BppChar fill = 0;
			if (argc == 2) fill = (BppChar) (BppInt) bp[2];
			BppCharBuffer* b = new BppCharBuffer((unsigned) size);
			*bp = b;
			for (BppUInt i=0;i<size;++i) b->push_back(fill);
		}
		break;
	default:
		THROW(_T("invalid argument type"));
		break;
	}
	sp = bp;
	return true;
}

// dict = newdict([bool sorted = false])
bool BppDefaultFunctions::Bpp_newdict(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	ARGCMAX(1);
	bool sorted = false;
	if (argc == 1)
		sorted = (((BppInt) *sp) != 0);
	sp -= argc;
	*sp = new BppDictionary(sorted);
	return true;
}

bool BppDefaultFunctions::Bpp_clone(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	BppValue v(*sp,true);
	*(--sp) = BPPMOVE(v);
	return true;
}

#ifdef __MSDOS__

#ifndef __TURBOC__
#define  coreleft _memmax
#endif
// size = memsize()
bool BppDefaultFunctions::Bpp_memsize(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	ARGCMAX(0);
		  *sp = (BppInt) coreleft();
	return true;
}

#endif

	// I/O functions
// fp = fopen(name,mode)
bool BppDefaultFunctions::Bpp_fopen(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(2,2);
	ValPtr bp = sp;
	bp-=2;
	BppString* filename = bp[1];
	BppString* mode = bp[2];
	if ((filename == 0) || (mode == 0))
		THROW(_T("invalid argument type"));
	FILE *fp = _tfopen(filename->c_str(),mode->c_str());
	if (fp != 0)
		*bp = fp;
	else
		bp->dispose();
	sp = bp;
	return true;
}

// done = fclose(fp)
bool BppDefaultFunctions::Bpp_fclose(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	if (sp->getType() != BVT_FILE)
		THROW(_T("invalid argument type"));
	 sp[-1] = (BppInt) fclose(*sp);
	 --sp;
	 return true;
}

// end = feof(fp)
bool BppDefaultFunctions::Bpp_feof(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	if (sp->getType() != BVT_FILE)
		THROW(_T("invalid argument type"));
	 sp[-1] = (BppInt) feof((FILE*)*sp);
	 --sp;
	 return true;
}

// ch = getc([fp])
bool BppDefaultFunctions::Bpp_getc(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	ARGCMAX(1);
	if (argc == 1)
	{
		if (sp->getType() != BVT_FILE)
			THROW(_T("invalid argument type"));
		ValPtr bp = sp; --bp;
		BppValue* pv = bp;
		BppInt l = (BppInt) getc((FILE*)(*sp));
		*pv = l;
		sp=bp;
	}
	else
		*sp = (BppInt) vm->getCurrentInput()->getChar();
	 return true;
}

// ch = putc(ch,fp)
bool BppDefaultFunctions::Bpp_putc(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(2,2);
	ValPtr bp = sp; bp -= argc;
	FILE* fp = bp[2];
	if ((!bp[1].isIntegralType()) || (fp == 0))
		THROW(_T("invalid argument type"));
	 *bp = (BppInt) putc((int)(BppInt)bp[1],fp);
	 sp = bp;
	 return true;
}
#if !(defined(WINCE) || defined(__linux__))
#ifndef _gettche // watcom hack
	#define _gettche getche
#endif
#endif

// str = gets()
bool BppDefaultFunctions::Bpp_gets(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	// CHKVM;
	 BppString* s = new BppString();
	*s = vm->getCurrentInput()->getString();
	*sp = s;
	return true;
}

// str = fgets(fp)
bool BppDefaultFunctions::Bpp_fgets(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	FILE* fp = *sp;
	if (fp == 0)
		THROW(_T("invalid argument type"));
	BppChar c = 0;
	BppString* s = new BppString();
	while (((c = (BppChar) getc(fp)) != _T('\n')) &&(!feof(fp)))
	{
		(*s) += c;
	}
	--sp;
	*sp = s;
	return true;
}

// result = fputs(str,fp)
bool BppDefaultFunctions::Bpp_fputs(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(2,2);
	ValPtr bp = sp; bp -= argc;
	FILE* fp = bp[2];
	BppString* s = bp[1];
	if ((s == 0) || (fp == 0))
		THROW(_T("invalid argument"));
	*bp = (BppInt) fputs(s->a_str(),fp);
	 sp = bp;
	 return true;
}


// var = freadval(fp)
bool BppDefaultFunctions::Bpp_freadval(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	FILE* fp = *sp;
	if (fp == 0)
		THROW(_T("invalid argument type"));
	BppDictionaryIO io(*vm);
	--sp;
	if (!io.readSingleValue(fp,sp[0]))
		sp[0].dispose();
	return true;
}

// ok = fwriteval(value, fp)
bool BppDefaultFunctions::Bpp_fwriteval(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(2,2);
	ValPtr bp = sp; bp -= argc;
	FILE* fp = bp[2];
	if (fp == 0)
		THROW(_T("invalid argument type"));
	BppDictionaryIO io(*vm);
	*bp = (BppInt) ((io.writeSingleValue(fp,bp[1])) ? 1 : 0);
	sp = bp;
	return true;
}

// int fread(buffer,[[unsigned size=1,]unsigned cnt,]FILE* fp)
bool BppDefaultFunctions::Bpp_fread(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(2,4);
	ValPtr bp = sp; bp -= argc;
	FILE* fp = bp[argc];
	if (fp == 0) THROW(_T("invalid argument type"),_T(" (FILE expected)"));
	BppBuffer* pbuf = bp[1];
	if (pbuf == 0) THROW(_T("invalid argument type"),_T(" (Buffer expected)"));
	size_t cnt = 0;
	bool fixedsize = (argc > 2) ? true : false;
	if (fixedsize)
	{
		size_t size = 1;
		BppValue& vcnt = bp[argc-1];
		if (!vcnt.isIntegralType())
			THROW(_T("invalid argument type"),_T(" (int expected)"));
		cnt = (size_t) (BppUInt) vcnt;
		if (argc == 4)
		{
			BppValue& vsize = bp[2];
			if (!vsize.isIntegralType())
				THROW(_T("invalid argument type"),_T(" (int expected)"));
			size = (size_t) (BppUInt) vsize;
	  }
		size_t bsize = pbuf->size();
		size_t requestedSize = bsize + (cnt * size);
		if (requestedSize > pbuf->capacity())
			pbuf->setCapacity(requestedSize);
		cnt = fread(&(pbuf->data()[bsize]),size,cnt,fp);
		pbuf->setSize(bsize + (cnt * size));
	}
	else
	{
		int c;
		while ((c = fgetc(fp)) != EOF)
		{
			pbuf->push_back((BppByte) c);
			++cnt;
		}
	}
	*bp = (BppUInt) cnt;
	sp = bp;
	return true;
}


// int fwrite(buffer,[[unsigned size=1,]unsigned cnt,]FILE* fp)
bool BppDefaultFunctions::Bpp_fwrite(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(2,4);
	ValPtr bp = sp; bp -= argc;
	FILE* fp = bp[argc];
	if (fp == 0) THROW(_T("invalid argument type"),_T(" (FILE expected)"));
	BppBuffer* pbuf = bp[1];
	if (pbuf == 0) THROW(_T("invalid argument type"),_T(" (Buffer expected)"));
	size_t size = 1;
	size_t cnt = 0;
	bool fixedsize = (argc > 2) ? true : false;
	if (fixedsize)
	{
		BppValue& vcnt = bp[argc-1];
		if (!vcnt.isIntegralType())
			THROW(_T("invalid argument type"),_T(" (int expected)"));
		cnt = (size_t) (BppUInt) vcnt;
		if (argc == 4)
		{
			BppValue& vsize = bp[2];
			if (!vsize.isIntegralType())
				THROW(_T("invalid argument type"),_T(" (int expected)"));
			size = (size_t) (BppUInt) vsize;
	  }
		cnt = fwrite(pbuf->data(),size,cnt,fp);
	}
	else
	{
		cnt = pbuf->size();
		cnt = fwrite(pbuf->data(),size,cnt,fp);
	}
	*bp = (BppUInt) cnt;
	sp = bp;
	return true;
}

#if (INT_MAX > SHRT_MAX)
bool BppDefaultFunctions::Bpp_getwc(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	if (sp->getType() != BVT_FILE)
		THROW(_T("invalid argument type"));
	 ValPtr bp = sp; --bp;
	 BppValue* pv = bp;
	 BppInt l = (BppInt) fgetwc((FILE*)(*sp));
	 *pv = l;
	 sp=bp;
	 return true;
}

bool BppDefaultFunctions::Bpp_putwc(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(2,2);
	ValPtr bp = sp; bp -= argc;
	FILE* fp = bp[2];
	if ((!bp[1].isIntegralType()) || (fp == 0))
		THROW(_T("invalid argument type"));
	 *bp = (BppInt) fputwc((wchar_t)(BppUInt)bp[1],fp);
	 sp = bp;
	 return true;
}

bool BppDefaultFunctions::Bpp_fgetws(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	FILE* fp = *sp;
	if (fp == 0)
		THROW(_T("invalid argument type"));
	wchar_t c = 0;
	BppString* s = new BppString();
	while (((c = (BppChar) fgetwc(fp)) != L'\n') &&(!feof(fp)))
	{
		(*s) += (BppChar) c;
	}
	--sp;
	*sp = s;
	return true;
}

bool BppDefaultFunctions::Bpp_fputws(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(2,2);
	ValPtr bp = sp; bp -= argc;
	FILE* fp = bp[2];
	BppString* s = bp[1];
	if ((s == 0) || (fp == 0))
		THROW(_T("invalid argument"));
	*bp = (BppInt) fputws(s->w_str(),fp);
	 sp = bp;
	 return true;
}
#endif

bool BppDefaultFunctions::Bpp_print(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	ValPtr bp = sp; bp -= argc;
	for (int i=1;i<=argc;i++)
	{
		BppValue& v = bp[i];
		switch (v.getType())
		{
		case BVT_INT:
	  case BVT_SHORT:
	  case BVT_SBYTE:
#if (INT_MAX == SHRT_MAX)
		  vm->printer()(_T("%ld"),(BppInt) v);
#else
		  vm->printer()(_T("%d"),(BppInt) v);
#endif
			break;
	  case BVT_UINT:
	  case BVT_USHORT:
	  case BVT_BYTE:
#if (INT_MAX == SHRT_MAX)
		  vm->printer()(_T("%lu"),(BppUInt) v);
#else
		  vm->printer()(_T("%u"),(BppUInt) v);
#endif
			break;
	  case BVT_INTPTR:
			vm->printer()(_T("%p"),(BppIntPtr) v);
		  break;
	  case BVT_LONG:
#ifdef EMULATE_I64
		  {
			  BppChar buf[65];
			  ((BppLong)v).toa(buf);
			  vm->printer()(buf);
		  }
#else
			vm->printer()(_T("%lld"),(BppLong) v);
#endif
		 break;
	  case BVT_ULONG:
#ifdef EMULATE_I64
		  {
			  char buf[65];
			  ((BppULong)v).toa(buf);
			  vm->printer()(buf);
		  }
#else
			vm->printer()(_T("%llu"),(BppULong) v);
#endif
		  break;
		case BVT_FLOAT:
		case BVT_DOUBLE:
			vm->printer()(_T("%g"),(double) v);
			break;
		default:
		{
			BppProcInfo* pi = interpreter->getProcInfo();
			BppClass* ownClass = pi ? pi->getClass() : 0;
			BppString* s = interpreter->getStringFromValue(ownClass,v);
			if (s != 0)
				vm->printer()(_T("%s"), s->c_str());
			else
				vm->printer()(_T("(%s:%p)"), getValueTypeName(v.getType()), v.ptr());
			break;
		}
		}
	}
	sp = bp;
	sp->dispose();
	return true;
}

	// string functions
// len = strlen(s)
bool BppDefaultFunctions::Bpp_strlen(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	BppString* s = *sp;
	if (s == 0)
		THROW(_T("invalid argument type"));
	--sp;
	*sp = (BppUInt) s->length();
	return true;
}

// result = strcmp(s1,s2)
bool BppDefaultFunctions::Bpp_strcmp(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(2,2);
	ValPtr bp = sp; bp -= argc;
	BppString* s1 = bp[1];
	BppString* s2 = bp[2];
	if ((s1 == 0) || (s2 == 0))
		THROW(_T("invalid argument type"));
	*bp = (BppInt) s1->compare(*s2);
	sp = bp;
	return true;
}

// result = stricmp(s1,s2)
bool BppDefaultFunctions::Bpp_stricmp(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(2,2);
	ValPtr bp = sp; bp -= argc;
	BppString* s1 = bp[1];
	BppString* s2 = bp[2];
	if ((s1 == 0) || (s2 == 0))
		THROW(_T("invalid argument type"));
	*bp = (BppInt) s1->icompare(*s2);
	sp = bp;
	return true;
}

// vec = strsplit(str,delimit,storeempty=false,storedelimit=false)
bool BppDefaultFunctions::Bpp_strsplit(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(2,4);
	ValPtr bp = sp; bp -= argc;
	BppString* s1 = bp[1];
	BppString* s2 = bp[2];
	if ((s1 == 0) || (s2 == 0))
		THROW(_T("invalid argument type"));
	BppInt emptytok = argc > 2 ? (BppInt)bp[3] : 0;
	BppInt storedelim = argc > 3 ? (BppInt)bp[4] : 0;
	BppVector* vec = new BppVector();
	s1->split(*vec, *s2, emptytok != 0, storedelim != 0);
	vec->shrink();
	*bp = vec;
	sp = bp;
	return true;
}

// int pos = strfind(string str,string s, uint offset=0);
// int pos = strfind(string str, long c, int offset = 0)
bool BppDefaultFunctions::Bpp_strfind(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(2,3);
	ValPtr bp = sp; bp -= argc;
	BppString* str = bp[1];
	if (str == 0)
	  THROW(_T("invalid argument type - string expected"));
	BppString* s = bp[2];
	BppChar c = (s == 0) ? (BppChar) (BppInt) bp[2] : 0;
	if ((s == 0) && (c == 0))
		THROW(_T("invalid argument - string or character expected"));
	size_t pos = (argc == 3) ? (size_t) (BppUInt) bp[3] : 0;
	if (s != 0)
	  *bp = (BppInt) str->find(*s,pos);
	else
		*bp = (BppInt) str->find(c,pos);
	sp = bp;
	return true;
}

// string modifiedstring = strerase(string str, uint pos, int cnt=-1)
// modifies str
bool BppDefaultFunctions::Bpp_strerase(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(2,3);
	ValPtr bp = sp; bp -= argc;
	BppString* str = bp[1];
	if (str == 0)
		THROW(_T("invalid argument type - string expected"));
	size_t pos = (size_t) (BppUInt) bp[2];
	long cnt = (argc == 3) ? (BppInt) bp[3] : -1;
	str->erase(pos,cnt);
	*bp = str;
	sp = bp;
	return true;
}
// string modifiedstring = strinsert(string str, uint pos, string insertion)
// string modifiedstring = strinsert(string str, uint pos , int c, uint cnt=1)
bool BppDefaultFunctions::Bpp_strinsert(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(3,4);
	ValPtr bp = sp; bp -= argc;
	BppString* str = bp[1];
	if (str == 0)
		THROW(_T("invalid argument type - string expected"));
	BppUInt pos = bp[2];
	BppString* s = bp[3];
	if (s != 0)
	{
		ARGCMAX(3);
		str->insert((size_t)pos,*s);
	}
	else
	{
		size_t cnt = (argc == 4) ? (size_t) (BppUInt) bp[4] : 1;
		str->insert((size_t) pos,(BppChar) (BppInt) bp[3],cnt);
	}
	*bp = str;
	sp = bp;
	return true;
}

// string modifiedstring = strreplace(string str, string s, string replacement)
// string modifiedstring = strreplace(string str, int c, int replacement)
bool BppDefaultFunctions::Bpp_strreplace(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(3,3);
	ValPtr bp = sp; bp -= argc;
	BppString* str = bp[1];
	if (str == 0)
		THROW(_T("invalid argument type - string expected"));
	BppString* s = bp[2];
	if (s != 0)
	{
		BppString* rep = bp[3];
		if (rep == 0)
			THROW(_T("invalid argument type - string expected"));
		str->replace(*s,*rep);
	}
	else
	{
		str->replace((BppChar)(BppInt)bp[2],(BppChar)(BppInt)bp[3]);
	}
	*bp = str;
	sp = bp;
	return true;
}

bool BppDefaultFunctions::Bpp_struplow(BppVirtualMachine * vm, int argc, ValPtr & sp, bool upper)
{
	CHKVM;
	CHKARGC(1,1);
		ValPtr bp = sp; bp -= argc;
	BppString* str = *sp;
	if (str == 0)
		THROW(_T("invalid argument type - string expected"));
	--sp;
	
	if (upper) str->upper();
	else str->lower();
	*sp = str;
	return true;
}

bool BppDefaultFunctions::Bpp_strupr(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	return Bpp_struplow(vm, argc, sp, true);
}

bool BppDefaultFunctions::Bpp_strlwr(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	return Bpp_struplow(vm, argc, sp, false);
}

// string sub = substr(string str,uint start, int len=-1)
bool BppDefaultFunctions::Bpp_substr(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(2,3);
	ValPtr bp = sp; bp -= argc;
	BppString* str = bp[1];
	if (str == 0)
		THROW(_T("invalid argument type - string expected"));
	BppUInt pos = bp[2];
	BppInt len = (argc == 3) ? (BppInt) bp[3] : -1;
	*bp = new BppString(str->substr((size_t) pos,len));
	sp = bp;
	return true;
}

	// vector and buffer related functions
// uint size = vecsize(vec)
bool BppDefaultFunctions::Bpp_vecsize(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	BppVector* v = *sp;
	if (v == 0)
		THROW(_T("invalid argument type"));
	--sp;
	*sp = (BppUInt) v->size();
	return true;
}

// size = size(vec|buf|cbuf|dict)
bool BppDefaultFunctions::Bpp_size(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	size_t s = 0;
	BppValueType vtype = maskWeakRef(sp->getType());
	if (!sp->addr()) vtype = BVT_NULL;
	switch (vtype)
	{
	case BVT_VECTOR:
		s = ((BppVector*)(*sp))->size();
		break;
	case BVT_BUFFER:
		s = ((BppBuffer*)(*sp))->size();
		break;
	case BVT_CHARBUFFER:
		s = ((BppCharBuffer*)(*sp))->size();
		break;
	case BVT_DICTIONARY:
		s = ((BppDictionary*)(*sp))->count();
		break;
	case BVT_STRING:
		s = ((BppString*)(*sp))->length();
		break;
	default:
			  THROW(_T("invalid argument type"));
	}
	--sp;
	*sp = (BppUInt) s;
	return true;
}

// uint
bool BppDefaultFunctions::Bpp_shrink(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKARGC(1, 1);
	size_t s = 0;
	BppValueType vtype = maskWeakRef(sp->getType());
	if (!sp->addr()) vtype = BVT_NULL;
	switch (vtype)
	{
	case BVT_VECTOR:
		s = ((BppVector*)(*sp))->shrink();
		break;
	case BVT_BUFFER:
		s = ((BppBuffer*)(*sp))->shrink();
		break;
	case BVT_CHARBUFFER:
		s = ((BppCharBuffer*)(*sp))->shrink();
		break;
	default:
		THROW(_T("invalid argument type"));
	}
	--sp;
	*sp = (BppUInt)s;
	return true;
}

//int pos =  find(vec cont, var searchVal, int pos=0)
bool BppDefaultFunctions::Bpp_find(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(2,3);
	ValPtr bp = sp; bp -= argc;
	BppInt pos = 0;
	if (argc == 3) pos = bp[3];
	BppVector* pvec = bp[1];
	if (pvec) 
		*bp = (BppInt) pvec->find(bp[2],pos);
	else
	{
		THROW(_T("invalid argument type"));
	}
	sp = bp;
	return true;
}

//int pos =  rfind(vec cont, var searchVal, int pos=-1)
bool BppDefaultFunctions::Bpp_rfind(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(2,3);
	ValPtr bp = sp; bp -= argc;
	BppInt pos = -1;
	if (argc == 3) pos = bp[3];
	BppVector* pvec = bp[1];
	if (pvec) 
		*bp = (BppInt) pvec->rfind(bp[2],pos);
	else
	{
		THROW(_T("invalid argument type"));
	}
	sp = bp;
	return true;
}

#ifndef DOXY
static const BppString _OP_CALL = _T("OP_CALL");

// static qsort helpers
// variables are initialized by Bpp_qsort function
struct QSort {
// comparer callback function value - may be object, bytecode or native function
	BppValue cmp_func;
	BppObject* pobj;
	BppValue cmp_obj;
	int argcnt;
// temporary pointer to stack pointer
	ValPtr* cmpSP;
// temporary VM pointer
	BppVirtualMachine* pvm;
// flag for success, set by doCpmpare function
	bool cmpOk;
	BppVector args;
	BppValueType ftype;
	QSort(BppVirtualMachine* pVm, ValPtr* sp)
		: pobj(0), argcnt(2), cmpSP(sp), pvm(pVm), cmpOk(true), args(2,true),ftype(BVT_NULL) {}
	void setFunc(const BppValue& f) { 
		pobj = f;
		if (pobj) {
			cmp_obj = f;
			args.insert(0, cmp_obj);
			++argcnt;
			BppDictionaryEntry* pe = pobj->getClass()->findMember(0,_OP_CALL);
			if (pe) cmp_func = pe->getValue();
		}
		else cmp_func = f;
		ftype = cmp_func.getType();
	}
	// compares two values using cmp_func
	int doCompare(const BppValue& l, const BppValue& r)
	{
		if (ftype == BVT_NULL) return l.compare(r);
		BppValue result;
		cmpOk = false;
	  args[argcnt - 2] = l;
	  args[argcnt - 1] = r;
	  cmpOk = pvm->interpreter()->invokeFunc(result, cmp_func, &args);
	  return (BppInt)result;
	}
	int cmp(BppUInt l ,BppUInt r)
	{
		if (ftype == BVT_NULL)
			return l < r ? -1 : r < l ? 1 : 0;
		return doCompare(l,r);
	}
	inline void arrSwap(BppRefObject* arr, int i1, int i2, BppValueType vt) {
		if (i1 == i2) return;
		switch (vt)
		{
		case BVT_BUFFER:
		{
			BppByte* p = ((BppBuffer*)arr)->data();
			BppByte v = p[i1];
			p[i1] = p[i2]; p[i2] = v;
			return;
		}
		case BVT_CHARBUFFER:
		{
			BppChar* p = ((BppCharBuffer*)arr)->data();
			BppChar v = p[i1];
			p[i1] = p[i2]; p[i2] = v;
			return;
		}
		}
		((BppVector*)arr)->swap(i1,i2);
	}
	void qusort(BppBuffer& arr, int i1, int i2)
	{
		if (i1 == i2) return;
		int l = i1;
		int r = i2;
		BppByte p = arr[(i1 + i2) >> 1 ];
		do {
			while (cmpOk && cmp(arr[r],p) > 0) --r;
			while (cmpOk && cmp(arr[l],p) < 0) ++l;
			if ((!cmpOk) || (l > r)) break;
			arrSwap(&arr,r, l,BVT_BUFFER);
		} while (++l <= --r);
		if (cmpOk && (i1 < r)) qusort(arr, i1, r);
		if (cmpOk && (l < i2)) qusort(arr, l, i2);
	}
	void qusort(BppCharBuffer& arr, int i1, int i2)
	{
		if (i1 == i2) return;
		int l = i1;
		int r = i2;
		BppChar p = arr[(i1 + i2) >> 1];
		do {
			while (cmpOk && cmp(arr[r], p) > 0) --r;
			while (cmpOk && cmp(arr[l], p) < 0) ++l;
			if ((!cmpOk) || (l > r)) break;
			arrSwap(&arr, r, l, BVT_CHARBUFFER);
		} while (++l <= --r);
		if (cmpOk && (i1 < r)) qusort(arr, i1, r);
		if (cmpOk && (l < i2)) qusort(arr, l, i2);
	}
	void qusort(BppVector& arr, int i1, int i2)
	{
		if (i1 == i2) return;
		if (i1 + 1 == i2) {
			if (doCompare(arr[i1], arr[i2]) > 0) {
				if (cmpOk) arr.swap(i1, i2);
			}
			return;
		}
		int l = i1;
		int r = i2;
		BppValue  p = arr[(i1 + i2) >> 1];
		do {
			while (cmpOk && doCompare(arr[r], p) > 0) --r;
			while (cmpOk && doCompare(arr[l], p) < 0) ++l;
			if ((!cmpOk) || (l > r)) break;
			arr.swap(r, l);
		} while (++l <= --r);
		if (r - i1 < i2 - l) 
		{
			if (cmpOk && (i1 < r)) qusort(arr, i1, r);
			if (cmpOk && (l < i2)) qusort(arr, l, i2);
		}
		else
		{
			if (cmpOk && (l < i2)) qusort(arr, l, i2);
			if (cmpOk && (i1 < r)) qusort(arr, i1, r);
		}
	}
};



#endif 
// null qsort(var container, var comparer)
bool BppDefaultFunctions::Bpp_qsort(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,2);
	ValPtr bp = sp; bp -= argc;
	QSort _sorter(vm, &sp);
	if (argc == 2) _sorter.setFunc(bp[2]);
	BppValueType vt = maskWeakRef(bp[1].getType());
	switch (vt)
	{
	case BVT_VECTOR:
		{
			BppVector* vec = bp[1];
			  if (!vec) THROW(_T("invalid container"));
			  if (vec->size() > 0) _sorter.qusort(*vec, 0, vec->size() - 1);
		}
		break;
	case BVT_BUFFER:
		{
			BppBuffer* buf = bp[1];
			  if (!buf) THROW(_T("invalid container"));
			  if (buf->size() > 0) _sorter.qusort(*buf, 0, buf->size() - 1);
		}
		break;
	case BVT_CHARBUFFER:
		{
			BppCharBuffer* buf = bp[1];
			if (!buf) THROW(_T("invalid container"));
			if (buf->size() > 0) _sorter.qusort(*buf, 0, buf->size() - 1);
		}
		break;
	default:
		THROW(_T("invalid container type"));
	}
	sp = bp;
	sp->dispose();
	return _sorter.cmpOk;
}

// uint foreach(var container, var func, ...)
bool BppDefaultFunctions::Bpp_foreach(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	ARGCMIN(2);
	ValPtr bp = sp; bp -= argc;
	BppValue vfunc = bp[2];
	BPPUSERFUNC pufunc = vfunc;
	BppObject* pfuncobj = pufunc ? 0 : (BppObject*)vfunc;
	BppVector* pfunc = 0;
	if (!(pufunc || pfuncobj)) {
		if (maskWeakRef(vfunc.getType()) == BVT_BYTECODE)
			pfunc = vfunc;
		if (!pfunc) THROW(_T("invalid function argument"));
	}
	BppValue& vcont = bp[1];
	BppValue resval;
	BppObject* contobj = vcont;
	if (contobj) {
		BppVector fargs(1, true);
		fargs.addRef();
		fargs[0] = vfunc;
		for (int i = 3; i <= argc; ++i) fargs.push_back(bp[i]);
		if (!interpreter->invokeMember(resval, contobj, _T("foreach"), &fargs)) return false;
		*bp = BPPMOVE(resval);
		sp = bp;
		return true;
	}
	BppValueType vt = maskWeakRef(vcont.getType());
	int argoff = 0;
	BppDictionaryEntry* method = 0;
	//const BppChar* opname = _T("OP_CALL");
	if (pfuncobj) {
		method = pfuncobj->getClass()->findMember(0, _OP_CALL);
		if (method) {
			++argoff;
			//opname = 0;
			vfunc = method->getValue();
		}
	}
	BppVector fargs(argc+argoff,true);
	fargs.addRef();
	if (method) fargs[0] = pfuncobj;//  method->getValue();
	for (int i = 3; i <= argc; ++i) fargs[argoff+i-1] = (const BppValue&) bp[i];
	bool ok = true;
	BppUInt loopcnt = 0;
	switch (vt)
	{
	case BVT_VECTOR:
	{
		BppVector* vec = vcont;
		if (!vec) THROW(_T("invalid container"));
		unsigned n = vec->size();
		for (BppUInt i = 0; i < n; ++i) {
			fargs[argoff] = i;
			fargs[argoff+1] = (*vec)[i];
			//if (pfuncobj)
			//   ok = interpreter->invokeMember(resval, pfuncobj, opname, &fargs);
			//else
				ok = interpreter->invokeFunc(resval, vfunc, &fargs);
			if (!ok) return false;
			++loopcnt;
			if (!(BppInt)resval) break;
		}
	}
	break;
	case BVT_DICTIONARY:
	{
		BppDictionary* dict = vcont;
		if (!dict) THROW(_T("invalid container"));
		for (BppDictionaryEntry* i = dict->getFirst();i;i=i->getNext())
		{
			fargs[argoff] = i->getKey();
			fargs[argoff+1] = i->getValue();
			//if (pfuncobj)
			//   ok = interpreter->invokeMember(resval, pfuncobj, opname, &fargs);
			//else
				ok = interpreter->invokeFunc(resval, vfunc, &fargs);
			if (!ok) return false;
			++loopcnt;
			if (!(BppInt)resval) break;
		}
	}
	break;
	case BVT_BUFFER:
	{
		BppBuffer* buf = vcont;
		if (!buf) THROW(_T("invalid container"));
		size_t n = buf->size();
		for (BppUInt i = 0; i < n; ++i) {
			fargs[argoff] = i;
			fargs[argoff+1] = (*buf)[i];
			//if (pfuncobj)
			//   ok = interpreter->invokeMember(resval, pfuncobj, opname, &fargs);
			//else
				ok = interpreter->invokeFunc(resval, vfunc, &fargs);
			if (!ok) return false;
			++loopcnt;
			if (!(BppInt)resval) break;
		}
	}
	break;
	case BVT_CHARBUFFER:
	{
		BppCharBuffer* buf = vcont;
		size_t n = buf->size();
		for (BppUInt i = 0; i < n; ++i) {
			fargs[argoff] = i;
			fargs[argoff+1] = (*buf)[i];
			//if (pfuncobj)
			//   ok = interpreter->invokeMember(resval, pfuncobj, opname, &fargs);
			//else
				ok = interpreter->invokeFunc(resval, vfunc, &fargs);
			if (!ok) return false;
			++loopcnt;
			if (!(BppInt)resval) break;
		}
	}
	break;
	case BVT_STRING:
	{
		BppString* buf = vcont;
		size_t n = buf->length();
		for (BppUInt i = 0; i < n; ++i) {
			fargs[argoff] = i;
			fargs[argoff+1] = (*buf)[i];
			//if (pfuncobj)
			//   ok = interpreter->invokeMember(resval, pfuncobj, opname, &fargs);
			//else
				ok = interpreter->invokeFunc(resval, vfunc, &fargs);
			if (!ok) return false;
			++loopcnt;
			if (!(BppInt)resval) break;
		}
	}
	break;
	default:
		THROW(_T("invalid container type"));
	}
	sp = bp;
	*bp = loopcnt;
	return true;
}

// vector resvec = vinsert(vector vec, int pos, var insertval, uint cnt = 1)
bool BppDefaultFunctions::Bpp_vinsert(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(3,4);
	ValPtr bp = sp; bp -= argc;
	BppVector* vec = bp[1];
	if (!vec)
			THROW(_T("invalid argument type"));
	BppInt pos = bp[2];
	BppUInt cnt = (argc == 4) ?  (BppInt) bp[4] : 1;
	vec->insert(pos,bp[3],cnt);
	*bp = vec;
	sp = bp;
	return true;
}

bool BppDefaultFunctions::Bpp_vswap(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(3,3);
	ValPtr bp = sp; bp -= argc;
	BppVector* vec = bp[1];
	if (!vec)
			THROW(_T("invalid argument type"));
	BppUInt pos1 = bp[2];
	BppUInt pos2 = bp[3];
	size_t s = vec->size();
	if (pos1 >= s || pos2 >= s) THROW (_T("index out of range"));
	vec->swap(pos1,pos2);
	*bp = vec;
	sp = bp;
	return true;
}

bool BppDefaultFunctions::Bpp_vmove(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(3,3);
	ValPtr bp = sp; bp -= argc;
	BppVector* vec = bp[1];
	if (!vec)
			THROW(_T("invalid argument type"));
	BppUInt pos1 = bp[2];
	BppUInt pos2 = bp[3];
	size_t s = vec->size();
	if (pos1 > s || pos2 > s) THROW (_T("index out of range"));
	vec->move(pos1,pos2);
	*bp = vec;
	sp = bp;
	return true;
}


// vector resvec = verase(vector vec, uint pos = 0, int cnt = -1)
bool BppDefaultFunctions::Bpp_verase(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,3);
	ValPtr bp = sp; bp -= argc;
	BppVector* vec = bp[1];
	if (!vec)
			THROW(_T("invalid argument type"));
	BppUInt pos = (argc > 1) ? (BppUInt) bp[2] : (BppUInt)0;
  BppInt cnt = (argc == 3) ? (BppInt) bp[3] : -1;
	vec->erase(pos,cnt);
	*bp = vec;
	sp = bp;
	return true;
}

// vector resvec = vslice(vector vec, uint from = 0, cnt = -1)
bool BppDefaultFunctions::Bpp_vslice(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,3);
	ValPtr bp = sp; bp -= argc;
	BppVector* vec = bp[1];
	if (!vec)
			THROW(_T("invalid argument type"));
	BppUInt pos = (argc > 1) ? (BppUInt) bp[2] : (BppUInt)0;
	 BppInt cnt = (argc == 3) ? (BppInt) bp[3] : (BppInt)-1;
	*bp = new BppVector(vec->slice(pos,cnt));
	sp = bp;
	return true;
}

// void vclear(vector vec)
bool BppDefaultFunctions::Bpp_vclear(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	ValPtr bp = sp; bp -= argc;
	BppVector* vec = bp[1];
	if (!vec)
		THROW(_T("invalid argument type"));
	vec->clear();
	bp->dispose();
	sp = bp;
	return true;
}


// bool found = dictcontains(dict,string key)
bool BppDefaultFunctions::Bpp_dictcontains(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(2,2);
	ValPtr bp = sp; bp -= argc;
	BppDictionary* dict = bp[1];
	BppString* key = bp[2];
	if ((dict == 0) || (key == 0))
		THROW(_T("invalid argument type"));
	*bp = (BppInt) ((dict->find(*key) != 0) ? 1 : 0);
	sp = bp;
	return true;
}

// bool deleted = dicterase(dict,string key)
bool BppDefaultFunctions::Bpp_dicterase(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(2,2);
	ValPtr bp = sp; bp -= argc;
	BppDictionary* dict = bp[1];
	BppString* key = bp[2];
	if ((dict == 0) || (key == 0))
		THROW(_T("invalid argument type"));
	*bp = (BppInt) (dict->erase(*key) ? 1 : 0);
	sp = bp;
	return true;
}

// null dictclear(dict)
bool BppDefaultFunctions::Bpp_dictclear(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	BppDictionary* dict = *sp;
	if (dict == 0)
		THROW(_T("invalid argument type"));
	dict->clear();
	--sp;
	sp->dispose();
	return true;
}

// keyvec = dictgetkeys(dict)
bool BppDefaultFunctions::Bpp_dictgetkeys(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	BppDictionary* dict = *sp;
	if (dict == 0)
		THROW(_T("invalid argument type"));
	BppVector* result = new BppVector();
		for (BppDictionaryEntry* pe = dict->getFirst(); pe != 0; pe = pe->getNext())
			result->push_back(new BppString(*(pe->getKey())));
	--sp;
	*sp = result;
	return true;
}

// keyvec = dictgetvalues(dict)
bool BppDefaultFunctions::Bpp_dictgetvalues(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	BppDictionary* dict = *sp;
	if (dict == 0)
		THROW(_T("invalid argument type"));
	BppVector* result = new BppVector();
		for (BppDictionaryEntry* pe = dict->getFirst(); pe != 0; pe = pe->getNext())
		result->push_back(pe->getValue());
	--sp;
	*sp = result;
	return true;
}

// v = Vec(...)
bool BppDefaultFunctions::Bpp_Vec(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	return Bpp_T(vm,argc,sp);
}

// v = T(...)
bool BppDefaultFunctions::Bpp_T(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	//CHKVM;
	ValPtr bp = sp; bp -= argc;
	BppVector* v = new BppVector((unsigned)argc);
	for(int i=1;i<=argc;i++)
		v->push_back(bp[i]);
	*bp = v;
	sp = bp;
	return true;
}

// cb = CBuf(...) // argtypes: integral, buffer, charbuffer, string; other ignored
bool BppDefaultFunctions::Bpp_CBuf(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
  // CHKVM;
	ValPtr bp = sp; bp -= argc;
	BppCharBuffer* cb = new BppCharBuffer((unsigned)argc);
	for(int i=1;i<=argc;i++)
	{
		BppValue& v = bp[i];
		BppValueType vt = maskWeakRef(v.getType());
		switch (vt)
		{
	  case BVT_BYTE:
	  case BVT_SBYTE:
			cb->push_back((BppChar) (BppByte) v);
		 break;
	  case BVT_SHORT:
	  case BVT_USHORT:
	  case BVT_INT:
	  case BVT_UINT:
	  case BVT_LONG:
	  case BVT_ULONG:
	  case BVT_INTPTR:
			cb->push_back((BppChar) (BppUShort) v);
		 break;
		case BVT_BUFFER:
			{
				BppBuffer* b = v;
				if (b != 0)
				{
					size_t n = b->size();
					for (size_t j=0; j<n; j++)
						cb->push_back((BppChar)(*b)[j]);
				}
			}
			break;
		case BVT_STRING:
			{
				BppString* s = v;
				if (s != 0)
				{
					size_t n =  s->length();
					for (size_t j=0; j<n; j++)
						cb->push_back((BppChar)((*s)[j]));
				}
			}
			break;

		case BVT_CHARBUFFER:
			{
				BppCharBuffer* pcb = v;
				if (pcb != 0)
					(*cb) += (*pcb);
			}
			break;
		}
	}
	*bp = cb;
	sp = bp;
	return true;
}


	// program related functions
// ok = compile(string src | FILE* fp)
bool BppDefaultFunctions::Bpp_compile(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,3);
	ValPtr bp = sp; bp -= argc;
	BppValue& v = bp[1];
	const BppChar* srcName = _T("");
	BppInt debug = 0;
	if (argc == 3)
	{
		debug = bp[3];
	}
	if (argc >1)
	{
		BppString* s = bp[2];
		if (s == 0)
			 THROW(_T("invalid argument type"));
		srcName = s->c_str();
	}
	bool ok = false;
	switch (v.getType())
	{
	case BVT_STRING:
	case BVT_STRINGREF:
		{
			BppString* s = v;
			if (s == 0) break;
			BppCompiler comp(*vm);
			if (debug) {
				comp.setGenerateLineInfo(true);
				comp.setGenerateLocalVarInfo(true);
			}
			ok = comp.compile(s->c_str(),srcName);
		}
		break;
	case BVT_FILE:
		{
			BppCompiler comp(*vm);
			if (debug) {
				comp.setGenerateLineInfo(true);
				comp.setGenerateLocalVarInfo(true);
			}
			ok = comp.compile((FILE*) v, srcName);
		}
		break;
	default:
		THROW(_T("invalid argument type"));
		break;
	}
	*bp = (BppInt) (ok ? 1 : 0);
	sp = bp;
	return true;
}

	// program related functions
// ok = compilefn(string src,string args="",string rtype="",var ownerClass = null, string srcname = "")
bool BppDefaultFunctions::Bpp_compilefn(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,6);
	ValPtr bp = sp; bp -= argc;
	bool argok = true;
	const BppChar* srcName = _T("");
	const BppChar* rtype = _T("");
	const BppChar* params = _T("");
	BppString* s = 0;
	BppClass* pc = 0;
	BppInt debug = false;
	switch (argc)
	{
	case 6:
		debug = bp[6];
	case 5: // source name
		s = bp[5];
		if (!s) { argok = false; break; }
		srcName = s->c_str();
	case 4: // class or class name
		{
			BppValue& v = bp[4];
			switch (v.getType())
			{
			case BVT_CLASS:
			case BVT_CLASSREF:
				pc = v;
				break;
			case BVT_STRING:
			case BVT_STRINGREF:
				{
					s = v;
					if (!s) { argok = false; break; }
					if (!s->length()) break;
					BppDictionaryEntry* pe = vm->classes().find(*s);
					if (!pe) THROW(_T("unknown class name "),s->c_str());
					pc = pe->getValue();
				}
				break;
			case BVT_NULL:
				break;
			default:
				argok = false;
			}
		}
		if (!argok) break;
	case 3: // result type
		s = bp[3];
		if (!s) { argok = false; break; }
		rtype = s->c_str();
	case 2: // formal parameter list
		s = bp[2];
		if (!s) { argok = false; break; }
		params = s->c_str();
	}
	s = bp[1];
	if (!(s && argok)) THROW(_T("invalid argument type"));
	BppCompiler comp(*vm);
	if (debug) {
		comp.setGenerateLineInfo(true);
		comp.setGenerateLocalVarInfo(true);
	}
	*bp = BppValue(comp.compileFunction(s->c_str(),params,rtype,srcName,pc),true);
	sp = bp;
	return true;
}


bool BppDefaultFunctions::Bpp_litexpr(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKARGC(1, 2);
	ValPtr bp = sp; bp -= argc;
	BppString* s = bp[1];
	BppString* sn = 0;
	if (!s) THROW(_T("invalid argument type"));
	if (argc > 1) {
		sn = bp[2];
		if (!sn) THROW(_T("invalid argument type"));
	}
	BppStringPrinter pr;
	BppCompiler comp(*vm,&pr);
	bool ok =  comp.litExpr(s->c_str(),*bp,sn ? sn->c_str():_T(""));
	sp = bp;
	if (!ok) THROW(pr.c_str());
	return ok;
}

// ok = loadmodule(filename)
bool BppDefaultFunctions::Bpp_loadmodule(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	BppString* filename = *sp;
	if (filename == 0)
		THROW(_T("invalid argument type"));
	BppDictionaryIO dict(*vm);
	bool ok = dict.readModule(*filename);
	--sp;
	*sp = (BppInt) (ok ? 1 : 0);
	return true;
}


	// basic type conversion
#ifndef DOXY
static const BppString s_bool = _T("bool");
static const BppString s_int = _T("int");
static const BppString s_buffer = _T("buffer");
static const BppString s_charbuffer = _T("charbuffer");
static const BppString s_vector = _T("vector");
static const BppString s_dictionary = _T("dictionary");
#endif
// int result = bool(val)
bool BppDefaultFunctions::Bpp_bool(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	BppValue& v = *sp;
	ValPtr bp = sp;
	bp -= argc;
	bool result = false;
	BppValueType vt = maskWeakRef(v.getType());
	if (v.isIntegralType()) {
		if ( vt >=
#ifdef BPP_I64
			BVT_INTPTR
#else
			BVT_LONG
#endif
	  ) result = (BppULong(v) != BppULong(0));
		else result = (BppUInt(v) != 0);
	}
	else if (v.isRealType()) result = ((double)v != 0.0);
	else {
		result = (v.ptr() != 0);
		if (vt == BVT_OBJECT || vt == BVT_STRONGOBJECT)
		{
			BppObject*pobj = v;
			if (pobj != 0)
			{
				BppProcInfo* pi = interpreter->getProcInfo();
				BppClass* ownClass = pi ? pi->getClass() : 0;
				BppClass* pclass = pobj->getClass();
				BppDictionaryEntry* pe = pclass->findMember(ownClass, s_bool);
				if ((pe == 0) || (pe->getSymbolType() != ST_FUNCTION))
					pe = pclass->findMember(ownClass, s_int);
				if ((pe != 0) && (pe->getSymbolType() == ST_FUNCTION))
				{
					*(++sp) = pe->getValue();
					if (interpreter->send(1, 0))
					{
						result = ((BppInt)*sp) == 0 ? 0 : 1;
					}
				}
			} // if pobj
		}
	}
	sp = bp;
	*sp = (BppInt)( result ? 1 : 0);
	return true;
}



bool BppDefaultFunctions::numcast(BppVirtualMachine* vm, int argc, ValPtr& sp, const BppChar* fname, BppValueType desttype)
{
	CHKVM;
	CHKARGC(1, 1);
	ValPtr bp = sp;
	const BppValue& src = *bp;
	BppValue& dest = * (--bp);
	bool ok = interpreter->doNumCast(dest, src, desttype, fname);
	sp = bp;
	 return ok;
}

bool BppDefaultFunctions::Bpp_sbyte(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	return numcast(vm, argc, sp, _T("sbyte"), BVT_SBYTE);
}

bool BppDefaultFunctions::Bpp_byte(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	return numcast(vm, argc, sp, _T("byte"), BVT_BYTE);
}

bool BppDefaultFunctions::Bpp_short(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	return numcast(vm, argc, sp, _T("short"), BVT_SHORT);
}

bool BppDefaultFunctions::Bpp_ushort(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	return numcast(vm, argc, sp, _T("ushort"), BVT_USHORT);
}

bool BppDefaultFunctions::Bpp_int(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	return numcast(vm, argc, sp, _T("int"), BVT_INT);
}

bool BppDefaultFunctions::Bpp_uint(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	return numcast(vm, argc, sp, _T("uint"), BVT_UINT);
}

bool BppDefaultFunctions::Bpp_long(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	return numcast(vm, argc, sp, _T("long"), BVT_LONG);
}

bool BppDefaultFunctions::Bpp_ulong(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	return numcast(vm, argc, sp, _T("ulong"), BVT_ULONG);
}

bool BppDefaultFunctions::Bpp_intptr(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	return numcast(vm, argc, sp, _T("intptr"), BVT_INTPTR);
}



// result = float(numericval|string)
bool BppDefaultFunctions::Bpp_float(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	return numcast(vm, argc, sp, _T("float"), BVT_FLOAT);
}

// result = double(numericval|string)
bool BppDefaultFunctions::Bpp_double(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	return numcast(vm, argc, sp, _T("double"), BVT_DOUBLE);
}

// result = string(int|double|float|string|charbuffer|object)
bool BppDefaultFunctions::Bpp_string(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	const BppValue& src = *sp;
	ValPtr bp = sp;
	--bp;
	bool ok = interpreter->doStrCast(*bp, src);
	sp = bp;
	return ok;
}

// result = buffer(int|double|float|buffer|charbuffer|string|object)
bool BppDefaultFunctions::Bpp_buffer(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	BppValue& v = *sp;
	ValPtr bp = sp;
	--bp;
	BppValueType vt = maskWeakRef(v.getType());
	switch (vt)
	{
	case BVT_SBYTE:
	case BVT_BYTE:
		{
			BppBuffer* b = new BppBuffer(sizeof(BppByte));
			b->push_back(v);
			*bp = b;
		}
		break;
	case BVT_SHORT:
	case BVT_USHORT:
		{
			BppBuffer* b = new BppBuffer(sizeof(BppUShort));
			b->push_backW((BppUShort)v);
			*bp = b;
		}
		break;
	case BVT_INT:
	case BVT_UINT:
		{
			BppBuffer* b = new BppBuffer(sizeof(BppUInt));
			b->push_backDW((BppUInt)v);
			*bp = b;
		}
		break;
	case BVT_INTPTR:
#ifndef BPP_I64
		{
			BppBuffer* b = new BppBuffer(sizeof(BppUInt));
			b->push_backDW((BppUInt)v);
			*bp = b;
		 break;
		}
#endif
	case BVT_LONG:
	case BVT_ULONG:
	case BVT_DOUBLE:
		{
			BppBuffer* b = new BppBuffer(sizeof(double));
		 BppByte* pd = (BppByte*) &v._v;
		 for (int i = 0; i < sizeof(double); ++i)
			 b->push_back(pd[i]);
		  *bp = b;
		}
		break;
	case BVT_FLOAT:
		{
			BppBuffer* b = new BppBuffer(sizeof(float));
		 float f = v;
			b->push_backDW(*((BppDWord*)&f));
			*bp = b;
		}
		break;
	case BVT_BUFFER:
		*bp = v;
		break;
	case BVT_STRING:
		{
			BppString* s = v;
			if (s != 0)
			{
				size_t bytes = (s->length()+1) * sizeof(BppChar);
				BppBuffer* b = new BppBuffer(bytes);
			memcpy(b->data(),s->c_str(),bytes);
			b->setSize(bytes);
				*bp = b;
			}
			else
				*bp = new BppBuffer();
		}
		break;
	case BVT_CHARBUFFER:
		{
			BppCharBuffer* cb = v;
			if (cb != 0)
			{
			 size_t bytes = (cb->size()+1) * sizeof(BppChar);
				BppBuffer* b = new BppBuffer(bytes);
			BppChar c = 0;
			memcpy(b->data(),cb->data(),bytes-sizeof(BppChar));
			memcpy(b->data()+(bytes-sizeof(BppChar)),&c,sizeof(BppChar));
			b->setSize(bytes);
			  *bp = b;
			}
			else
				*bp = new BppBuffer();
		}
		break;
	case BVT_OBJECT:
	case BVT_STRONGOBJECT:
	{
			BppObject*pobj = v;
			if (pobj != 0)
			{
				BppClass* pclass = pobj->getClass();
				BppProcInfo* pi = interpreter->getProcInfo();
				BppClass* ownClass = pi ? pi->getClass() : 0;
				BppDictionaryEntry* pe = pclass->findMember(ownClass,s_buffer);
				if ((pe != 0) && (pe->getSymbolType() == ST_FUNCTION))
				{
				*(++sp) = pe->getValue();
					if (interpreter->send(1,0))
					{
						*bp = *sp;
				  vt = bp->getType();
				  if (vt == BVT_BUFFER || vt == BVT_BUFFERREF)
						  break;
					}
				}
			} // if pobj
		}
		default:
		 sp = bp;
		THROW(_T("invalid argument type"));
		break;
	}
	sp = bp;
	return true;
}

// result = buffer(int|double|float|buffer|charbuffer|string|object)
bool BppDefaultFunctions::Bpp_charbuffer(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	BppValue& v = *sp;
	ValPtr bp = sp;
	--bp;
	BppValueType vt = maskWeakRef(v.getType());
	switch (vt)
	{
	case BVT_INT:
	case BVT_SHORT:
	case BVT_SBYTE:
		{
#if (INT_MAX == SHRT_MAX)
			BppCharBuffer* b = new BppCharBuffer(BppString::format(_T("%ld"),(BppInt)v).c_str());
#else
			BppCharBuffer* b = new BppCharBuffer(BppString::format(_T("%d"),(BppInt)v).c_str());
#endif
			*bp = b;
		}
		break;
	case BVT_UINT:
	case BVT_USHORT:
	case BVT_BYTE:
		{
#if (INT_MAX == SHRT_MAX)
			BppCharBuffer* b = new BppCharBuffer(BppString::format(_T("%lu"),(BppUInt)v).c_str());
#else
			BppCharBuffer* b = new BppCharBuffer(BppString::format(_T("%u"),(BppUInt)v).c_str());
#endif
			*bp = b;
		}
		break;
	case BVT_INTPTR:
		{
			BppCharBuffer* b = new BppCharBuffer(BppString::format(_T("%p"),(BppIntPtr)v).c_str());
			*bp = b;
		}
		break;
	case BVT_LONG:
		{
#ifdef EMULATE_I64
			BppChar buf[65];
			BppCharBuffer* b = new BppCharBuffer((const BppChar*)((BppLong)v).toa(buf));
#else
			BppCharBuffer* b = new BppCharBuffer(BppString::format(_T("%lld"),(BppLong)v).c_str());
#endif
			*bp = b;
		}
		break;
	case BVT_ULONG:
		{
#ifdef EMULATE_I64
			BppChar buf[65];
			BppCharBuffer* b = new BppCharBuffer((const BppChar*)((BppULong)v).toa(buf));
#else
			BppCharBuffer* b = new BppCharBuffer(BppString::format(_T("%llu"),(BppULong)v).c_str());
#endif
			*bp = b;
		}
		break;
	case BVT_FLOAT:
	case BVT_DOUBLE:
		{
			BppCharBuffer* b = new BppCharBuffer(BppString::format(_T("%g"),(double)v).c_str());
			*bp = b;
		}
		break;
	case BVT_CHARBUFFER:
		*bp = v;
		break;
	case BVT_STRING:
		{
			BppString* s = bp[1];
			if (s != 0)
			{
				BppCharBuffer* b = new BppCharBuffer(s->c_str());
				*bp = b;
			}
			else
				*bp = new BppCharBuffer();
		}
		break;
	case BVT_BUFFER:
		{
			BppBuffer* b = bp[1];
			if (b != 0)
			{
				size_t size = b->size();
				BppCharBuffer* cb = new BppCharBuffer(size+1);
				for (size_t i=0;i<size;i++)
					cb->push_back((BppChar) (*b)[i]);
			cb->push_back(0);
				*bp = cb;
			}
			else
				*bp = new BppCharBuffer();
		}
		break;
	case BVT_OBJECT:
	case BVT_STRONGOBJECT:
		{
			BppObject*pobj = v;
			if (pobj != 0)
			{
				BppClass* pclass = pobj->getClass();
				BppProcInfo* pi = interpreter->getProcInfo();
				BppClass* ownClass = pi ? pi->getClass() : 0;
				BppDictionaryEntry* pe = pclass->findMember(ownClass,s_charbuffer);
				if ((pe != 0) && (pe->getSymbolType() == ST_FUNCTION))
				{
				*(++sp) = pe->getValue();
					if (interpreter->send(1,0))
					{
						*bp = *sp;
				  vt = bp->getType();
				  if (vt == BVT_CHARBUFFER || vt == BVT_CHARBUFFERREF)
						  break;
					}
				}
			} // if pobj
		}
		default:
		 sp = bp;
		THROW(_T("invalid argument type"));
		break;
	}
	sp = bp;
	return true;
}

// result = vector(vector|buffer|charbuffer|dictionary|object)
bool BppDefaultFunctions::Bpp_vector(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	BppValue& v = *sp;
	ValPtr bp = sp;
	--bp;
	BppValueType vt = maskWeakRef(v.getType());
	switch (vt)
	{
	case BVT_VECTOR:
		*bp = v;
		break;
	case BVT_DICTIONARY:
		return Bpp_dictgetvalues(vm,argc,sp);
	case BVT_CHARBUFFER:
		{
			BppCharBuffer* b = v;
			if (b)
			{
			  size_t size = b->size();
			  BppVector* vec = new BppVector(size,size ? true : false);
			  for (size_t i=0;i<size;++i)
				  (*vec)[i] = (*b)[i];
				  *bp = vec;
			}
			else *bp = new BppVector();
		}
		break;
	case BVT_BUFFER:
		{
			BppBuffer* b = v;
			if (b)
			{
			  size_t size = b->size();
			  BppVector* vec = new BppVector(size,size ? true : false);
			  for (size_t i=0;i<size;++i)
				  (*vec)[i] = (*b)[i];
				  *bp = vec;
			}
			else *bp = new BppVector();
		}
		break;
	case BVT_OBJECT:
	case BVT_STRONGOBJECT:
	{
			BppObject*pobj = v;
			if (pobj != 0)
			{
				BppClass* pclass = pobj->getClass();
				BppProcInfo* pi = interpreter->getProcInfo();
				BppClass* ownClass = pi ? pi->getClass() : 0;
				BppDictionaryEntry* pe = pclass->findMember(ownClass,s_vector);
				if ((pe != 0) && (pe->getSymbolType() == ST_FUNCTION))
				{
				*(++sp) = pe->getValue();
					if (interpreter->send(1,0))
					{
						*bp = *sp;
				  vt = bp->getType();
				  if (vt == BVT_VECTOR || vt == BVT_VECTORREF)
						  break;
					}
				}
			} // if pobj
		}
		default:
		 sp = bp;
		THROW(_T("invalid argument type"));
		break;
	}
	sp = bp;
	return true;
}

// result = vector(vector|dictionary|object)
bool BppDefaultFunctions::Bpp_dictionary(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	BppValue& v = *sp;
	ValPtr bp = sp;
	--bp;
	BppValueType vt = maskWeakRef(v.getType());
	switch (vt)
	{
	case BVT_DICTIONARY:
		*bp = v;
		break;
	case BVT_VECTOR:
		{
			BppDictionary* dict = new BppDictionary();
			*bp = dict;
			BppVector* vec = v;
			if (vec)
			{
			  size_t size = vec->size();
			  for (size_t i=0;i<size;++i)
			  {
				  (*dict)[BppString::format(_T("%d"),i)] = (*vec)[i];
			  }
				  *bp = dict;
			}
		}
		break;
	case BVT_OBJECT:
	case BVT_STRONGOBJECT:
		{
			BppObject*pobj = v;
			if (pobj != 0)
			{
				BppClass* pclass = pobj->getClass();
				BppProcInfo* pi = interpreter->getProcInfo();
				BppClass* ownClass = pi ? pi->getClass() : 0;
				BppDictionaryEntry* pe = pclass->findMember(ownClass,s_dictionary);
				if ((pe != 0) && (pe->getSymbolType() == ST_FUNCTION))
				{
				*(++sp) = pe->getValue();
					if (interpreter->send(1,0))
					{
						*bp = *sp;
				  vt = bp->getType();
				  if (vt == BVT_DICTIONARY || vt == BVT_DICTIONARYREF)
						  break;
					}
				}
			} // if pobj
		}
		default:
		 sp = bp;
		THROW(_T("invalid argument type"));
		break;
	}
	sp = bp;
	return true;
}


	// RTTI
// name = gettypename(typeid)
bool BppDefaultFunctions::Bpp_gettypename(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,2);
	ValPtr bp = sp; bp -= argc;
	BppValue& v = bp[1];
	if (!v.isIntegralType())
		  THROW(_T("invalid argument type"));
	bool prefix = false;
	if (argc > 1)
	{
		prefix = ((BppInt)bp[2]) != 0;
	}
	BppInt vt = v;
	if ((vt<_BVTMIN)|| (vt>_BVTREFMAX))
		  THROW(_T("invalid argument value"));
	sp = bp;
	*sp = new BppString(getValueTypeName((BppValueType) vt,prefix));
	return true;
}

bool BppDefaultFunctions::Bpp_gettype(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	BppValue& v = *sp;
	--sp;
	*sp = (BppInt) v.getType();
	return true;
}

bool BppDefaultFunctions::Bpp_getclassname(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	BppObject* o = *sp;
	if (o == 0)
		  THROW(_T("invalid argument type"));
	--sp;
	*sp = o->getClass()->getClassName();
	return true;
}

bool BppDefaultFunctions::Bpp_dynamic_cast(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(2,2);
	ValPtr bp = sp; bp -= 2;
	BppClass* pclass = bp[1];
	if (!pclass)
	{
		BppString* s = bp[1];
		if (s != 0)
		{
			BppDictionaryEntry* de = vm->classes().find(*s);
			if (de != 0)
				pclass = de->getValue();
		}
	}
	if (!pclass)
	{
		bp->dispose();
		sp = bp;
		THROW(_T("first argument is not a class"));
	}
	BppObject* pobj = bp[2];
	if (!pobj)
		bp->dispose();
	else
	{
		if (pobj->getClass()->isSubtypeOf(pclass))
			*bp = bp[2];
		else
			bp->dispose();
	}
	sp = bp;
	return true;
}

bool BppDefaultFunctions::Bpp_getsymbol(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKARGC(1, 2);
	ValPtr bp = sp; bp -= argc;
	BppString* pname = bp[1];
	if (!pname) THROW(_T("first argument is not a string"));
	bool findclass = false;
	if (argc == 2) {
		const BppValue& v = bp[2];
		if (!v.isIntegralType()) return interpreter->badType(v.getType(), BVT_INT);
		findclass = ((BppInt)v) != 0;
	}
	const BppDictionary& table = findclass ? vm->classes() : vm->symbols();
	BppDictionaryEntry* pe = table.find(*pname);
	if (pe) *bp = pe->getValue();
	else bp->dispose();
	sp = bp;
	return true;
}


// intval = time();
bool BppDefaultFunctions::Bpp_time(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	ARGCMAX(0);
#ifdef EMULATE_I64
	*sp =  (BppInt)time(0);
#else
	*sp = (BppLong)time(0);
#endif
	return true;
}

#ifndef CLK_TCK 
#define CLK_TCK CLOCKS_PER_SEC
#endif
// int mselapsed = timer()
bool BppDefaultFunctions::Bpp_timer(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	ARGCMAX(0);
#ifdef WINCE
	*sp = (BppUInt) GetTickCount();
#else
	*sp = (BppUInt) ((clock() * 1000L) / CLK_TCK);
#endif
	return true;
}

#ifdef _WIN32_WCE_EMULATION
  typedef double (__cdecl *mathfunc) (double);
#else
  typedef double (*mathfunc) (double);
#endif

#ifndef DOXY

static bool mfunc(mathfunc mf, BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	double d = *sp;
	if (d == DBL_MAX) {
	  if (!interpreter->doNumCast(*sp,*sp,BVT_DOUBLE,_T("double")))
		  return false;
	 d = *sp;
	}
	--sp;
	*sp = mf(d);
	return true;
}
#endif

	// math functions
bool BppDefaultFunctions::Bpp_sin(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	return mfunc(&sin,vm,argc,sp);
}

bool BppDefaultFunctions::Bpp_cos(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	return mfunc(&cos,vm,argc,sp);
}

bool BppDefaultFunctions::Bpp_tan(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	return mfunc(&tan,vm,argc,sp);
}

bool BppDefaultFunctions::Bpp_atan(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	return mfunc(&atan,vm,argc,sp);
}

bool BppDefaultFunctions::Bpp_sqrt(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	return mfunc(&sqrt,vm,argc,sp);
}

bool BppDefaultFunctions::Bpp_log(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	return mfunc(&log,vm,argc,sp);
}

bool BppDefaultFunctions::Bpp_exp(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	return mfunc(&exp,vm,argc,sp);
}

bool BppDefaultFunctions::Bpp_abs(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(1,1);
	switch (sp->getType())
	{
	case BVT_BYTE:
	case BVT_USHORT:
	case BVT_UINT:
	case BVT_ULONG:
		sp[-1] = *sp;
		break;
	case BVT_SBYTE:
		{
			BppSByte x = *sp;
			sp[-1] = (BppSByte) (x<0 ? -x : x);
		}
		break;
	case BVT_SHORT:
		{
			BppShort x = *sp;
			sp[-1] = (BppShort) (x<0 ? -x : x);
		}
		break;
	case BVT_INT:
		{
			BppInt x = *sp;
			sp[-1] = x<0 ? -x : x;
		}
		break;
	case BVT_LONG:
		{
			BppLong x = *sp;
			sp[-1] = x<BppLong(0L) ? -x : x;
		}
		break;
	case BVT_INTPTR:
		{
			BppIntPtr x = *sp;
			sp[-1] = x<BppIntPtr(0L) ? -x : x;
		}
		break;
	case BVT_FLOAT:
		{
			float x = *sp;
			sp[-1] = x<0.0f ? -x : x;
		}
		break;
	case BVT_DOUBLE:
		{
			double x = *sp;
			sp[-1] = x<0.0 ? -x : x;
		}
		break;
	default:
		 THROW(_T("invalid argument type"));
	}
	--sp;
	return true;
}

bool BppDefaultFunctions::Bpp_pow(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKARGC(2,2);
	ValPtr bp = sp; bp -= 2;
	double arg1 = bp[1];
	if (arg1 == DBL_MAX)
	{
		BppValue v;
		if (!interpreter->doNumCast(v, bp[1], BVT_DOUBLE, _T("double")))
				return false;
		arg1 = v;
	}
	double arg2 = bp[2];
	if (arg2 == DBL_MAX)
	{
		BppValue v = bp[2];
		if (!interpreter->doNumCast(v, bp[2], BVT_DOUBLE, _T("double")))
			return false;
		arg2 = v;
	}
	*bp = (double)pow(arg1, arg2); // exp(arg2 * log(arg1));
	 sp = bp;
	 return true;
}


#endif // !DOXY



