
#include "bppufunc.h"
#ifdef __OLDCPP__
#include <dir.h>
#endif
#include <stdio.h>
#include <stdlib.h>

#ifndef _MAX_PATH
#if (defined(GCC) && !defined(WIN32))
	#define _MAX_PATH PATH_MAX
#else
 #define _MAX_PATH 255
#endif
#endif

void errorExit(int err)
{
	exit(err);
}

void* allocmem(size_t blksize)
{
	void* result = malloc(blksize);
	if (!result)
		errorExit(BP_ERR_NOMEM);
	return result;
}

void* reallocmem(void* p, size_t blksize)
{
	void* result = realloc(p,blksize);
	if (!result)
		errorExit(BP_ERR_NOMEM);
	return result;
}

BppChar* dupstr(const BppChar* src)
{
	BppChar* result = _tcsdup(src);
	if (!result)
		errorExit(BP_ERR_NOMEM);
	 return result;
}

void freemem(void* p)
{
	free(p);
}

BPPLIB_DLLEXP BppIntPtr movemem(BppIntPtr dst, BppIntPtr src, BppUInt n)
{
	if (n==0 || dst == src) return dst;
	if (n == 0) return dst;
#ifdef __MSDOS__
	char huge * _dst = (char huge*) dst;
	char huge * _src = (char huge*) src;
	if (dst > src) {
		for (_dst += n-1, _src += n-1; n > 0; --_dst, --_src, --n) *_dst = *_src;
	}
	else {
		for (; n > 0; ++_dst, ++_src, --n) *_dst = *_src;
	}
	return dst;
#else
	return INTPTRCAST(memmove((void*) dst, (const void*) src, n));
#endif
}



const BppChar* getFileDir(const BppChar* filePath, bool stripSlash)
{
	static BppChar buf[_MAX_PATH];
	BppChar* dirpath = _tfullpath(buf,filePath,_MAX_PATH);
	 if (!dirpath) return 0;
	 int i = _tcslen(dirpath)-2;
	for (;i>1;--i)
	{
		BppChar c = dirpath[i];
		if (c == _T('/') || c == _T('\\') || c == _T(':'))
		{
			dirpath[i+(stripSlash ? 0 : 1)] = _T('\0');
			break;
		}
	}
	 return buf;
}


const BppChar* searchPath(const BppChar* fname)
{
	if (fname == 0) return 0;
#ifdef __OLDCPP__
	return searchpath(fname);
#else
		const BppChar* result = searchEnv(_T("Path"),fname);
	  if (result == 0)
		  result = searchEnv(_T("PATH"),fname);
	  return result;
#endif
}

const BppChar* searchEnv(const BppChar* varname, const BppChar* fname)
{
  if ((varname == 0) || (fname == 0)) return 0;
#if defined(__OLDCPP__)  || (defined(GCC) && (!defined(WIN32)))
	BppChar* envval = getenv(varname);
	return searchDirList(envval,fname);
#else
  #ifndef WINCE
		static BppChar buf[_MAX_PATH +1];
		_tsearchenv(fname,varname,buf);
	  if (_tcslen(buf) == 0) return 0;
	  return buf;
  #else
	const BppChar* envval = _tgetenv(varname);
	return searchDirList(envval,fname);
  #endif // WINCE
#endif
}

const BppChar* searchDirList(const BppChar* dirList, const BppChar* fname)
{
#ifdef UNIX
#define DIRDELIMIT _T("/")
#else
#define DIRDELIMIT _T("\\")
#endif
	static BppChar buf[_MAX_PATH +1];
	if ((dirList == 0) || (fname == 0)) return 0;
	//int dListLen = _tcslen(dirList);
	BppChar* dirbuf = dupstr(dirList);
//   _tcscpy(dirbuf,dirList);
	bool found = false;
	for (BppChar* token = _tcstok(dirbuf,_T(";"));(!found) && (token != 0);token = _tcstok(NULL,_T(";")))
	{
		_tcscpy(buf,token);
	  int buflen = _tcslen(buf);
	  if ((buflen > 0) && (buf[buflen-1] != _T('\\')) && (buf[buflen-1] != _T('/')))
		  _tcscat(buf,DIRDELIMIT);
		_tcscat(buf,fname);
		FILE* fp = _tfopen(buf,_T("r"));
		if (fp != 0)
		{
			fclose(fp);
			found = true;
		}
	}
	freemem(dirbuf);
	if (!found)
		return 0;
	return buf;
}

#ifdef WINCE
	int _gettche()
	{
		int result = _gettchar();
//      if (result)
//         _puttchar(result);
		return result;
	}


const BppChar* _tgetenv(const BppChar* varname)
{
	static BppChar buf[4096];
	unsigned long bufsize = sizeof(buf);
	unsigned long type;
	 buf[0] = 0;
	HKEY hkResult;
	 if (RegOpenKeyEx(HKEY_CURRENT_USER,_T("Environment"),0,0,&hkResult) == ERROR_SUCCESS)
	{
		 RegQueryValueEx(hkResult,varname,0,&type,(unsigned char*)buf,&bufsize);
		RegCloseKey(hkResult);
	}
	return buf;
}

int _tputenv(const BppChar* envstr)
{
	HKEY hkResult;
	DWORD dwDisp;
	BppChar* buf = dupstr(envstr);
	int result = 0;
	BppChar* p = _tcschr(buf, _T('='));
	if (p != 0)
	{
		 *p = 0;
		p++;
		 result = RegCreateKeyEx(HKEY_CURRENT_USER,_T("Environment"),0,_T("REG_SZ"),0,0,0,&hkResult,&dwDisp);
		if (result == 0)
		{
			if ((*p) == 0) // remove entry
			{
				result = RegDeleteValue(hkResult,buf);
			}
			else
			{
				result = RegSetValueEx(hkResult,buf,0,REG_SZ,(const unsigned char*)p,(_tcslen(p)+1) * sizeof(BppChar));
			}
		}
		RegCloseKey(hkResult);
	}
	else result = -1;
	 freemem(buf);
	return result;
}

union __FTIME
{
	FILETIME ft;
	unsigned __int64 fti;
};
static SYSTEMTIME __reftime = { 1970,1,4,1,0,0,0,0};

time_t time(time_t* t)
{
	static bool initialized = false;
	static __FTIME __refftime;
	if (!initialized)
	{
		SystemTimeToFileTime(&__reftime,&__refftime.ft);
		initialized = true;
	}
	SYSTEMTIME currentTime;
	GetSystemTime(&currentTime);
	__FTIME currentFTime;
	SystemTimeToFileTime(&currentTime,&currentFTime.ft);
	if (t != 0)
	{
	  *t = (time_t) ((currentFTime.fti - __refftime.fti) / 10000000);
	  return *t;
	}
	return (time_t) ((currentFTime.fti - __refftime.fti) / 10000000);
}

BppChar* _tctime(time_t* tt)
{
	static BppChar* __weekdays[] = {_T("Sun"),_T("Mon"),_T("Tue"),_T("Wed"),_T("Thu"),_T("Fri"),_T("Sat")};
	static BppChar* __months[] = {_T(""),_T("Jan"),_T("Feb"),_T("Mar"),_T("Apr"),_T("May"),_T("Jun"),
		_T("Jul"),_T("Aug"),_T("Sep"),_T("Oct"),_T("Nov"),_T("Dec")};
	static BppChar buf[50];
	static bool initialized = false;
	static __FTIME __refftime;
	if (!tt) return 0;
	if (!initialized)
	{
		SystemTimeToFileTime(&__reftime,&__refftime.ft);
		initialized = true;
	}
	SYSTEMTIME currentTime;
	__FTIME currentFTime;
	currentFTime.fti = *tt;
	currentFTime.fti *= 10000000;
	currentFTime.fti += __refftime.fti;
	FileTimeToLocalFileTime(&currentFTime.ft,&currentFTime.ft);
	if (FileTimeToSystemTime(&currentFTime.ft,&currentTime) == 0) return 0;
	_stprintf(buf,_T("%s %s %02u %02u:%02u:%02u %04u\n"),
		__weekdays[currentTime.wDayOfWeek],
		__months[currentTime.wMonth],
		(unsigned) currentTime.wDay,
		(unsigned) currentTime.wHour,
		(unsigned) currentTime.wMinute,
		(unsigned) currentTime.wSecond,
		(unsigned) currentTime.wYear);
	return (BppChar*) buf;
}

tm* localtime( const time_t* t)
{
	static tm _tm;
	static bool initialized = false;
	static __FTIME __refftime;
	if (!t) return 0;
	if (!initialized)
	{
		SystemTimeToFileTime(&__reftime,&__refftime.ft);
		initialized = true;
	}
	if (t == 0) return 0;
	SYSTEMTIME currentTime;
	__FTIME currentFTime;
	currentFTime.fti = ((*t) * 10000000) + __refftime.fti;
	 if (FileTimeToLocalFileTime(&currentFTime.ft,&currentFTime.ft) == 0) return 0;
	 if (FileTimeToSystemTime(&currentFTime.ft,&currentTime) == 0) return 0;
	_tm.tm_sec = currentTime.wSecond;
	_tm.tm_min = currentTime.wMinute;
	_tm.tm_hour = currentTime.wHour;
	_tm.tm_mday = currentTime.wDay;
	_tm.tm_mon = currentTime.wMonth;
	_tm.tm_year = currentTime.wYear;
	_tm.tm_wday = currentTime.wDayOfWeek;
	_tm.tm_yday = -1;
	_tm.tm_isdst = -1;
	return &_tm;
}

#endif // WINCE

#ifdef emulate_vsctprintf

#ifdef WINCE
	#define NULDEVICE "\\Temp\\NUL"
	#define NULDEVICEW L"\\Temp\\NUL"
#else
	#ifdef __MSDOS__
	  #define  NULDEVICE "NUL"
	  #define  NULDEVICEW "NUL"
  #else
	  #ifndef WIN32
		  #define NULDEVICE "/dev/null"
		  #define NULDEVICEW L"/dev/null"
	  #else
		 #define  NULDEVICE "NUL"
		 #define  NULDEVICEW L"NUL"
	  #endif
	#endif
#endif
	int _vscprintf(const char* format, va_list args)
	{
	  FILE* fp = fopen(NULDEVICE,"wt");
	  int result = vfprintf(fp,format,args);
	  fclose(fp);
	  return result;
	}
#ifdef WIN32
	int _vscwprintf(const wchar_t* format, va_list args)
	{
		FILE* fp = _wfopen(NULDEVICEW,L"wt");
		int result = vfwprintf(fp,format,args);
		fclose(fp);
		return result;
	}
#endif
#endif

#if defined(UNIX)
   char* _fullpath(char* absPath,const char* relPath, size_t)
   {
	   return realpath(relPath,absPath);
   }
#endif

