#ifndef INT64_H
#define INT64_H

class UInt64;

class Int64
{
	friend class UInt64;
public:
   Int64();
   ~Int64();
   Int64(long high, unsigned long low);
   Int64(const Int64& val);
   Int64(long val);
   Int64(unsigned long val);
   Int64(int val);
   Int64(unsigned int val);
   Int64(short val);
   Int64(unsigned short val);
   Int64(float val);
   Int64(double val);

   Int64& operator=(const Int64& val);
   Int64& operator++();
   const Int64 pinc(); // emulates postfix operator
   Int64& operator--();
   const Int64 pdec(); // emulates postfix operator
   
   Int64 operator-() const;
   Int64 operator~() const;
   Int64& operator+=(const Int64& val);
   Int64& operator+=(const UInt64& val);
  Int64& operator-=(const Int64& val);
   Int64& operator-=(const UInt64& val);
   Int64& operator*=(const Int64& val);
   Int64& operator*=(const UInt64& val);
   Int64& operator/=(const Int64& val);
   Int64& operator/=(const UInt64& val);
   Int64& operator%=(const Int64& val);
   Int64& operator%=(const UInt64& val);
   Int64& operator<<=(unsigned nbits);
   Int64& operator>>=(unsigned nbits);
   Int64& operator&=(const Int64& val);
   Int64& operator|=(const Int64& val);
   Int64& operator^=(const Int64& val);
   Int64 operator<<(unsigned nbits) const;
   Int64 operator>>(unsigned nbits) const;
   int compare(const Int64& cmp) const;
   operator __int64() const;
   char* toa(char* buf = 0, int base = 10) const;
   int toInt() const;
   unsigned int toUInt() const;
   long toLong() const;
   unsigned long toULong() const;
private:
   friend Int64 operator*(const Int64& l, const Int64& r);
   static bool _idiv(const Int64& dividend, const Int64& divisor, Int64& q, Int64& r);
   unsigned long _low;
   long _high;
};

Int64 operator+(const Int64& l, const Int64& r);
Int64 operator-(const Int64& l, const Int64& r);
Int64 operator*(const Int64& l, const Int64& r);
Int64 operator/(const Int64& l, const Int64& r);
Int64 operator%(const Int64& l, const Int64& r);
Int64 operator&(const Int64& l, const Int64& r);
Int64 operator|(const Int64& l, const Int64& r);
Int64 operator^(const Int64& l, const Int64& r);

class UInt64
{
	friend class Int64;
public:
   UInt64();
   ~UInt64();
   UInt64(unsigned long high, unsigned long low);
   UInt64(const UInt64& val);
   UInt64(const Int64& val);
   UInt64(long val);
   UInt64(unsigned long val);
   UInt64(int val);
   UInt64(unsigned int val);
   UInt64(short val);
   UInt64(unsigned short val);
   UInt64(float val);
   UInt64(double val);

   UInt64& operator=(const UInt64& val);
   UInt64& operator++();
   const UInt64 pinc(); // emulates postfix operator
   UInt64& operator--();
   const UInt64 pdec(); // emulates postfix operator
   
   UInt64 operator-() const;
   UInt64 operator~() const;
   //UInt64& operator+=(const Int64& val);
   UInt64& operator+=(const UInt64& val);
   //UInt64& operator-=(const Int64& val);
   UInt64& operator-=(const UInt64& val);
   //UInt64& operator*=(const Int64& val);
   UInt64& operator*=(const UInt64& val);
   //UInt64& operator/=(const Int64& val);
   UInt64& operator/=(const UInt64& val);
   //UInt64& operator%=(const Int64& val);
   UInt64& operator%=(const UInt64& val);
   UInt64& operator<<=(unsigned nbits);
   UInt64& operator>>=(unsigned nbits);
   UInt64& operator&=(const UInt64& val);
   UInt64& operator|=(const UInt64& val);
   UInt64& operator^=(const UInt64& val);
   UInt64 operator<<(unsigned nbits) const;
   UInt64 operator>>(unsigned nbits) const;
   int compare(const UInt64& cmp) const;
   operator unsigned __int64() const;
   char* toa(char* buf = 0, int base = 10) const;
   int toInt() const;
   unsigned int toUInt() const;
   long toLong() const;
   unsigned long toULong() const;
private:
   friend UInt64 operator*(const UInt64& l, const UInt64& r);
   static bool _div(const UInt64& dividend, const UInt64& divisor, UInt64& q, UInt64& r);
   unsigned long _low;
   unsigned long _high;
};

UInt64 operator+(const UInt64& l, const UInt64& r);
UInt64 operator-(const UInt64& l, const UInt64& r);
UInt64 operator*(const UInt64& l, const UInt64& r);
UInt64 operator/(const UInt64& l, const UInt64& r);
UInt64 operator%(const UInt64& l, const UInt64& r);
UInt64 operator&(const UInt64& l, const UInt64& r);
UInt64 operator|(const UInt64& l, const UInt64& r);
UInt64 operator^(const UInt64& l, const UInt64& r);

#endif
