// libdeelx.cpp
//
// DEELX Regular Expression Engine (v1.2)
//
// Copyright 2006 (c) RegExLab.com
// All Rights Reserved.
//
// http://www.regexlab.com/deelx/
//
// Author: ʷΰ (sswater shi)
// sswater@gmail.com
//

#include "..\deelx.h"
#include "libdeelx.h"

regexp_handle __stdcall regexp_create()
{
	return new CRegexpA;
}

regexp_handle __stdcall regexpw_create()
{
	return new CRegexpW;
}

result_handle __stdcall result_create()
{
	return new MatchResult(0);
}

void __stdcall regexp_free(regexp_handle handle)
{
	delete (CRegexpA*)handle;
}

void __stdcall regexpw_free(regexp_handle handle)
{
	delete (CRegexpW*)handle;
}

void __stdcall result_free(result_handle result)
{
	delete (MatchResult*)result;
}

void __stdcall regexp_compile(regexp_handle handle, const char * pattern, int flag)
{
	((CRegexpA*)handle)->Compile(pattern, flag);
}

void __stdcall regexpw_compile(regexp_handle handle, const unsigned short * pattern, int flag)
{
	((CRegexpW*)handle)->Compile(pattern, flag);
}

void __stdcall regexp_match(regexp_handle handle, const char * text, int startpos, result_handle result)
{
	(*((MatchResult*)result)) = ((CRegexpA*)handle)->Match(text, startpos);
}

void __stdcall regexpw_match(regexp_handle handle, const unsigned short * text, int startpos, result_handle result)
{
	(*((MatchResult*)result)) = ((CRegexpW*)handle)->Match(text, startpos);
}

int __stdcall regexp_replace(regexp_handle handle, const char * text, const char * replaceto, int startpos, int times, char * buffer, int bufsize)
{
	char * result = ((CRegexpA*)handle)->Replace(text, replaceto, startpos, times);
	int length = strlen(result);

	if(buffer != 0 && bufsize > 0)
	{
		strncpy(buffer, result, bufsize);
		if(length < bufsize) buffer[length] = 0;
	}

	delete [] result;
	return length;
}

int __stdcall regexpw_replace(regexp_handle handle, const unsigned short * text, const unsigned short * replaceto, int startpos, int times, unsigned short * buffer, int bufsize)
{
	unsigned short * result = ((CRegexpW*)handle)->Replace(text, replaceto, startpos, times);
	int length = wcslen(result);

	if(buffer != 0 && bufsize > 0)
	{
		wcsncpy(buffer, result, bufsize);
		if(length < bufsize) buffer[length] = 0;
	}

	delete [] result;
	return length;
}

int __stdcall regexp_namednumber(regexp_handle handle, const char * name)
{
	return ((CRegexpA*)handle)->GetNamedGroupNumber(name);
}

int __stdcall regexpw_namednumber(regexp_handle handle, const unsigned short * name)
{
	return ((CRegexpW*)handle)->GetNamedGroupNumber(name);
}

int __stdcall result_ismatched(result_handle result)
{
	return ((MatchResult*)result)->IsMatched();
}

int __stdcall result_start(result_handle result)
{
	return ((MatchResult*)result)->GetStart();
}

int __stdcall result_end(result_handle result)
{
	return ((MatchResult*)result)->GetEnd();
}

int __stdcall result_maxgroup(result_handle result)
{
	return ((MatchResult*)result)->MaxGroupNumber();
}

int __stdcall result_groupstart(result_handle result, int number)
{
	return ((MatchResult*)result)->GetGroupStart(number);
}

int __stdcall result_groupend(result_handle result, int number)
{
	return ((MatchResult*)result)->GetGroupEnd(number);
}
