// libdeelx.h
//
// DEELX Regular Expression Engine (v1.2)
//
// Copyright 2006 (c) RegExLab.com
// All Rights Reserved.
//
// http://www.regexlab.com/deelx/
//
// Author: ʷΰ (sswater shi)
// sswater@gmail.com
//

#ifndef __LIB_DEELX_H__
#define __LIB_DEELX_H__

// flags
#ifndef _REGEX_FLAGS_DEFINED
	enum REGEX_FLAGS
	{
		NO_FLAG        = 0,
		SINGLELINE     = 0x01,
		MULTILINE      = 0x02,
		GLOBAL         = 0x04,
		IGNORECASE     = 0x08,
		RIGHTTOLEFT    = 0x10,
		EXTENDED       = 0x20,
	};
	#define _REGEX_FLAGS_DEFINED
#endif

// exports
#ifdef LIBDEELX_EXPORTS
	#define DEELX_API __declspec( dllexport )
#else
	#define DEELX_API __declspec( dllimport )
#endif

// handle
typedef void * regexp_handle;
typedef void * result_handle;

// api
#ifdef __cplusplus
	extern "C" {
#endif

// load runtime
#ifndef LOAD_DEELX_RUNTIME

// create & free
DEELX_API regexp_handle __stdcall regexp_create ();
DEELX_API regexp_handle __stdcall regexpw_create();
DEELX_API result_handle __stdcall result_create ();
DEELX_API void          __stdcall regexp_free   (regexp_handle handle);
DEELX_API void          __stdcall regexpw_free  (regexp_handle handle);
DEELX_API void          __stdcall result_free   (result_handle result);

// compile & match
DEELX_API void __stdcall regexp_compile     (regexp_handle handle, const char * pattern, int flag);
DEELX_API void __stdcall regexpw_compile    (regexp_handle handle, const unsigned short * pattern, int flag);
DEELX_API void __stdcall regexp_match       (regexp_handle handle, const char * text   , int startpos, result_handle result);
DEELX_API void __stdcall regexpw_match      (regexp_handle handle, const unsigned short * text   , int startpos, result_handle result);
DEELX_API int  __stdcall regexp_replace     (regexp_handle handle, const char * text   , const char * replaceto, int startpos, int times, char * buffer, int bufsize);
DEELX_API int  __stdcall regexpw_replace    (regexp_handle handle, const unsigned short * text   , const unsigned short * replaceto, int startpos, int times, unsigned short * buffer, int bufsize);

// get result
DEELX_API int  __stdcall regexp_namednumber (regexp_handle handle, const char * name);
DEELX_API int  __stdcall regexpw_namednumber(regexp_handle handle, const unsigned short * name);
DEELX_API int  __stdcall result_ismatched   (result_handle result);
DEELX_API int  __stdcall result_start       (result_handle result);
DEELX_API int  __stdcall result_end         (result_handle result);
DEELX_API int  __stdcall result_maxgroup    (result_handle result);
DEELX_API int  __stdcall result_groupstart  (result_handle result, int number);
DEELX_API int  __stdcall result_groupend    (result_handle result, int number);

#else

// create & free
typedef regexp_handle (__stdcall * fn_regexp_create )();
typedef regexp_handle (__stdcall * fn_regexpw_create)();
typedef result_handle (__stdcall * fn_result_create )();
typedef void          (__stdcall * fn_regexp_free   )(regexp_handle handle);
typedef void          (__stdcall * fn_regexpw_free  )(regexp_handle handle);
typedef void          (__stdcall * fn_result_free   )(result_handle result);

// compile & match
typedef void (__stdcall * fn_regexp_compile     )(regexp_handle handle, const char * pattern, int flag);
typedef void (__stdcall * fn_regexpw_compile    )(regexp_handle handle, const unsigned short * pattern, int flag);
typedef void (__stdcall * fn_regexp_match       )(regexp_handle handle, const char * text   , int startpos, result_handle result);
typedef void (__stdcall * fn_regexpw_match      )(regexp_handle handle, const unsigned short * text   , int startpos, result_handle result);
typedef int  (__stdcall * fn_regexp_replace     )(regexp_handle handle, const char * text   , const char * replaceto, int startpos, int times, char * buffer, int bufsize);
typedef int  (__stdcall * fn_regexpw_replace    )(regexp_handle handle, const unsigned short * text   , const unsigned short * replaceto, int startpos, int times, unsigned short * buffer, int bufsize);

// get result
typedef int  (__stdcall * fn_regexp_namednumber )(regexp_handle handle, const char * name);
typedef int  (__stdcall * fn_regexpw_namednumber)(regexp_handle handle, const unsigned short * name);
typedef int  (__stdcall * fn_result_ismatched   )(result_handle result);
typedef int  (__stdcall * fn_result_start       )(result_handle result);
typedef int  (__stdcall * fn_result_end         )(result_handle result);
typedef int  (__stdcall * fn_result_maxgroup    )(result_handle result);
typedef int  (__stdcall * fn_result_groupstart  )(result_handle result, int number);
typedef int  (__stdcall * fn_result_groupend    )(result_handle result, int number);

#endif//DEELX_RUNTIME_LOAD

#ifdef __cplusplus
	}
#endif

#endif//__LIB_DEELX_H__
