/*
 * Copyright (c) 2001-2002 Daniel van Vugt <danv@users.sf.net>
 * This file is part of the Rex library.
 * See LICENSE.txt for license details.
 */

#include "rexprivate.h"


parseresult_t
ParseBound(const REXCHAR *chars, size_t len, int *min, int *max)
{
	int i;
	int *dest;
	bool_t havemax;

	*min = 0;
	*max = 0;
	havemax = FALSE;
	dest = min;

	i = 0;
	while (i < len)
	{
		REXCHAR c;

		c = chars[i];
		if ( (c >= '0') && (c <= '9') )
		{
			int newbound;
			
			newbound = (*dest * 10) + (c - '0');
			if (newbound > REX_DUP_MAX)
			{
				return ENCODE_PARSE_ERROR(REX_BOUND_TOO_BIG, i);
			}
			*dest = newbound;
			if (dest == max)
			{
				havemax = TRUE;
			}
		}
		else if (c == '}')
		{
			if (i == 0)
			{
				return ENCODE_PARSE_ERROR(REX_EMPTY_BOUND, i);
			}
			break;
		}
		else if ((c == ',') && (dest == min) && (i > 0))
		{
			dest = max;
		}
		else
		{
			return ENCODE_PARSE_ERROR(REX_INVALID_BOUND_CHAR, i);
		}
		
		i++;
	}

	if (havemax)
	{
		if ( *min > *max )
		{
			return ENCODE_PARSE_ERROR(REX_INVALID_BOUND_RANGE, i-1);
		}
	}
	else
	{
		*max = (dest == min) ? *min : -1;
	}

	return i;
}

